#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
import getpass
import logging
import socket
import time
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Union,
)
from urllib.parse import urlparse

from hive_metastore.ThriftHiveMetastore import Client
from hive_metastore.ttypes import (
    AlreadyExistsException,
    CheckLockRequest,
    EnvironmentContext,
    FieldSchema,
    InvalidOperationException,
    LockComponent,
    LockLevel,
    LockRequest,
    LockResponse,
    LockState,
    LockType,
    MetaException,
    NoSuchObjectException,
    SerDeInfo,
    StorageDescriptor,
    UnlockRequest,
)
from hive_metastore.ttypes import Database as HiveDatabase
from hive_metastore.ttypes import Table as HiveTable
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_exponential
from thrift.protocol import TBinaryProtocol
from thrift.transport import TSocket, TTransport

from pyiceberg.catalog import (
    EXTERNAL_TABLE,
    ICEBERG,
    LOCATION,
    METADATA_LOCATION,
    TABLE_TYPE,
    URI,
    MetastoreCatalog,
    PropertiesUpdateSummary,
)
from pyiceberg.exceptions import (
    CommitFailedException,
    NamespaceAlreadyExistsError,
    NamespaceNotEmptyError,
    NoSuchIcebergTableError,
    NoSuchNamespaceError,
    NoSuchPropertyException,
    NoSuchTableError,
    TableAlreadyExistsError,
    WaitingForLockException,
)
from pyiceberg.partitioning import UNPARTITIONED_PARTITION_SPEC, PartitionSpec
from pyiceberg.schema import Schema, SchemaVisitor, visit
from pyiceberg.serializers import FromInputFile
from pyiceberg.table import (
    CommitTableResponse,
    StagedTable,
    Table,
    TableProperties,
)
from pyiceberg.table.sorting import UNSORTED_SORT_ORDER, SortOrder
from pyiceberg.table.update import (
    TableRequirement,
    TableUpdate,
)
from pyiceberg.typedef import EMPTY_DICT, Identifier, Properties
from pyiceberg.types import (
    BinaryType,
    BooleanType,
    DateType,
    DecimalType,
    DoubleType,
    FixedType,
    FloatType,
    IntegerType,
    ListType,
    LongType,
    MapType,
    NestedField,
    PrimitiveType,
    StringType,
    StructType,
    TimestampType,
    TimestamptzType,
    TimeType,
    UnknownType,
    UUIDType,
)
from pyiceberg.utils.properties import property_as_bool, property_as_float

if TYPE_CHECKING:
    import pyarrow as pa


COMMENT = "comment"
OWNER = "owner"

# If set to true, HiveCatalog will operate in Hive2 compatibility mode
HIVE2_COMPATIBLE = "hive.hive2-compatible"
HIVE2_COMPATIBLE_DEFAULT = False

HIVE_KERBEROS_AUTH = "hive.kerberos-authentication"
HIVE_KERBEROS_AUTH_DEFAULT = False
HIVE_KERBEROS_SERVICE_NAME = "hive.kerberos-service-name"
HIVE_KERBEROS_SERVICE_NAME_DEFAULT = "hive"

LOCK_CHECK_MIN_WAIT_TIME = "lock-check-min-wait-time"
LOCK_CHECK_MAX_WAIT_TIME = "lock-check-max-wait-time"
LOCK_CHECK_RETRIES = "lock-check-retries"
DEFAULT_LOCK_CHECK_MIN_WAIT_TIME = 0.1  # 100 milliseconds
DEFAULT_LOCK_CHECK_MAX_WAIT_TIME = 60  # 1 min
DEFAULT_LOCK_CHECK_RETRIES = 4
DO_NOT_UPDATE_STATS = "DO_NOT_UPDATE_STATS"
DO_NOT_UPDATE_STATS_DEFAULT = "true"

logger = logging.getLogger(__name__)


class _HiveClient:
    """Helper class to nicely open and close the transport."""

    _transport: TTransport
    _ugi: list[str] | None

    def __init__(
        self,
        uri: str,
        ugi: str | None = None,
        kerberos_auth: bool | None = HIVE_KERBEROS_AUTH_DEFAULT,
        kerberos_service_name: str | None = HIVE_KERBEROS_SERVICE_NAME,
    ):
        self._uri = uri
        self._kerberos_auth = kerberos_auth
        self._kerberos_service_name = kerberos_service_name
        self._ugi = ugi.split(":") if ugi else None
        self._transport = self._init_thrift_transport()

    def _init_thrift_transport(self) -> TTransport:
        url_parts = urlparse(self._uri)
        socket = TSocket.TSocket(url_parts.hostname, url_parts.port)
        if not self._kerberos_auth:
            return TTransport.TBufferedTransport(socket)
        else:
            return TTransport.TSaslClientTransport(socket, host=url_parts.hostname, service=self._kerberos_service_name)

    def _client(self) -> Client:
        protocol = TBinaryProtocol.TBinaryProtocol(self._transport)
        client = Client(protocol)
        if self._ugi:
            client.set_ugi(*self._ugi)
        return client

    def __enter__(self) -> Client:
        """Make sure the transport is initialized and open."""
        if not self._transport.isOpen():
            try:
                self._transport.open()
            except (TypeError, TTransport.TTransportException):
                # Close the old transport before reinitializing to prevent resource leaks
                try:
                    self._transport.close()
                except Exception:
                    pass
                # reinitialize _transport
                self._transport = self._init_thrift_transport()
                self._transport.open()
        return self._client()  # recreate the client

    def __exit__(self, exctype: type[BaseException] | None, excinst: BaseException | None, exctb: TracebackType | None) -> None:
        """Close transport if it was opened."""
        if self._transport.isOpen():
            self._transport.close()


def _construct_hive_storage_descriptor(schema: Schema, location: str | None, hive2_compatible: bool = False) -> StorageDescriptor:
    ser_de_info = SerDeInfo(serializationLib="org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe")
    return StorageDescriptor(
        [
            FieldSchema(field.name, visit(field.field_type, SchemaToHiveConverter(hive2_compatible)), field.doc)
            for field in schema.fields
        ],
        location,
        "org.apache.hadoop.mapred.FileInputFormat",
        "org.apache.hadoop.mapred.FileOutputFormat",
        serdeInfo=ser_de_info,
    )


PROP_EXTERNAL = "EXTERNAL"
PROP_TABLE_TYPE = "table_type"
PROP_METADATA_LOCATION = "metadata_location"
PROP_PREVIOUS_METADATA_LOCATION = "previous_metadata_location"
DEFAULT_PROPERTIES = {TableProperties.PARQUET_COMPRESSION: TableProperties.PARQUET_COMPRESSION_DEFAULT}


def _construct_parameters(
    metadata_location: str, previous_metadata_location: str | None = None, metadata_properties: Properties | None = None
) -> dict[str, Any]:
    properties = {PROP_EXTERNAL: "TRUE", PROP_TABLE_TYPE: "ICEBERG", PROP_METADATA_LOCATION: metadata_location}
    if previous_metadata_location:
        properties[PROP_PREVIOUS_METADATA_LOCATION] = previous_metadata_location

    if metadata_properties:
        for key, value in metadata_properties.items():
            if key not in properties:
                properties[key] = str(value)

    return properties


def _annotate_namespace(database: HiveDatabase, properties: Properties) -> HiveDatabase:
    params = {}
    for key, value in properties.items():
        if key == COMMENT:
            database.description = value
        elif key == LOCATION:
            database.locationUri = value
        else:
            params[key] = value
    database.parameters = params
    return database


HIVE_PRIMITIVE_TYPES = {
    BooleanType: "boolean",
    IntegerType: "int",
    LongType: "bigint",
    FloatType: "float",
    DoubleType: "double",
    DateType: "date",
    TimeType: "string",
    TimestampType: "timestamp",
    TimestamptzType: "timestamp with local time zone",
    StringType: "string",
    UUIDType: "string",
    BinaryType: "binary",
    FixedType: "binary",
    UnknownType: "void",
}


class SchemaToHiveConverter(SchemaVisitor[str]):
    hive2_compatible: bool

    def __init__(self, hive2_compatible: bool):
        self.hive2_compatible = hive2_compatible

    def schema(self, schema: Schema, struct_result: str) -> str:
        return struct_result

    def struct(self, struct: StructType, field_results: list[str]) -> str:
        return f"struct<{','.join(field_results)}>"

    def field(self, field: NestedField, field_result: str) -> str:
        return f"{field.name}:{field_result}"

    def list(self, list_type: ListType, element_result: str) -> str:
        return f"array<{element_result}>"

    def map(self, map_type: MapType, key_result: str, value_result: str) -> str:
        # Key has to be primitive for Hive
        return f"map<{key_result},{value_result}>"

    def primitive(self, primitive: PrimitiveType) -> str:
        if isinstance(primitive, DecimalType):
            return f"decimal({primitive.precision},{primitive.scale})"
        elif self.hive2_compatible and isinstance(primitive, TimestamptzType):
            # Hive2 doesn't support timestamp with local time zone
            return "timestamp"
        else:
            return HIVE_PRIMITIVE_TYPES[type(primitive)]


class HiveCatalog(MetastoreCatalog):
    _client: _HiveClient

    def __init__(self, name: str, **properties: str):
        super().__init__(name, **properties)
        self._client = self._create_hive_client(properties)

        self._lock_check_min_wait_time = property_as_float(properties, LOCK_CHECK_MIN_WAIT_TIME, DEFAULT_LOCK_CHECK_MIN_WAIT_TIME)
        self._lock_check_max_wait_time = property_as_float(properties, LOCK_CHECK_MAX_WAIT_TIME, DEFAULT_LOCK_CHECK_MAX_WAIT_TIME)
        self._lock_check_retries = property_as_float(
            properties,
            LOCK_CHECK_RETRIES,
            DEFAULT_LOCK_CHECK_RETRIES,
        )

    @staticmethod
    def _create_hive_client(properties: dict[str, str]) -> _HiveClient:
        last_exception = None
        for uri in properties[URI].split(","):
            try:
                return _HiveClient(
                    uri,
                    properties.get("ugi"),
                    property_as_bool(properties, HIVE_KERBEROS_AUTH, HIVE_KERBEROS_AUTH_DEFAULT),
                    properties.get(HIVE_KERBEROS_SERVICE_NAME, HIVE_KERBEROS_SERVICE_NAME_DEFAULT),
                )
            except BaseException as e:
                last_exception = e
        if last_exception is not None:
            raise last_exception
        else:
            raise ValueError(f"Unable to connect to hive using uri: {properties[URI]}")

    def _convert_hive_into_iceberg(self, table: HiveTable) -> Table:
        properties: dict[str, str] = table.parameters
        if TABLE_TYPE not in properties:
            raise NoSuchPropertyException(
                f"Property table_type missing, could not determine type: {table.dbName}.{table.tableName}"
            )

        table_type = properties[TABLE_TYPE]
        if table_type.lower() != ICEBERG:
            raise NoSuchIcebergTableError(
                f"Property table_type is {table_type}, expected {ICEBERG}: {table.dbName}.{table.tableName}"
            )

        if prop_metadata_location := properties.get(METADATA_LOCATION):
            metadata_location = prop_metadata_location
        else:
            raise NoSuchPropertyException(f"Table property {METADATA_LOCATION} is missing")

        io = self._load_file_io(location=metadata_location)
        file = io.new_input(metadata_location)
        metadata = FromInputFile.table_metadata(file)
        return Table(
            identifier=(table.dbName, table.tableName),
            metadata=metadata,
            metadata_location=metadata_location,
            io=self._load_file_io(metadata.properties, metadata_location),
            catalog=self,
        )

    def _convert_iceberg_into_hive(self, table: Table) -> HiveTable:
        identifier_tuple = table.name()
        database_name, table_name = self.identifier_to_database_and_table(identifier_tuple, NoSuchTableError)
        current_time_millis = int(time.time() * 1000)

        return HiveTable(
            dbName=database_name,
            tableName=table_name,
            owner=table.properties[OWNER] if table.properties and OWNER in table.properties else getpass.getuser(),
            createTime=current_time_millis // 1000,
            lastAccessTime=current_time_millis // 1000,
            sd=_construct_hive_storage_descriptor(
                table.schema(),
                table.location(),
                property_as_bool(self.properties, HIVE2_COMPATIBLE, HIVE2_COMPATIBLE_DEFAULT),
            ),
            tableType=EXTERNAL_TABLE,
            parameters=_construct_parameters(metadata_location=table.metadata_location, metadata_properties=table.properties),
        )

    def _create_hive_table(self, open_client: Client, hive_table: HiveTable) -> None:
        try:
            open_client.create_table(hive_table)
        except AlreadyExistsException as e:
            raise TableAlreadyExistsError(f"Table {hive_table.dbName}.{hive_table.tableName} already exists") from e

    def _get_hive_table(self, open_client: Client, database_name: str, table_name: str) -> HiveTable:
        try:
            return open_client.get_table(dbname=database_name, tbl_name=table_name)
        except NoSuchObjectException as e:
            raise NoSuchTableError(f"Table does not exists: {table_name}") from e

    def create_table(
        self,
        identifier: str | Identifier,
        schema: Union[Schema, "pa.Schema"],
        location: str | None = None,
        partition_spec: PartitionSpec = UNPARTITIONED_PARTITION_SPEC,
        sort_order: SortOrder = UNSORTED_SORT_ORDER,
        properties: Properties = EMPTY_DICT,
    ) -> Table:
        """Create a table.

        Args:
            identifier: Table identifier.
            schema: Table's schema.
            location: Location for the table. Optional Argument.
            partition_spec: PartitionSpec for the table.
            sort_order: SortOrder for the table.
            properties: Table properties that can be a string based dictionary.

        Returns:
            Table: the created table instance.

        Raises:
            AlreadyExistsError: If a table with the name already exists.
            ValueError: If the identifier is invalid.
        """
        properties = {**DEFAULT_PROPERTIES, **properties}
        staged_table = self._create_staged_table(
            identifier=identifier,
            schema=schema,
            location=location,
            partition_spec=partition_spec,
            sort_order=sort_order,
            properties=properties,
        )
        database_name, table_name = self.identifier_to_database_and_table(identifier)

        self._write_metadata(staged_table.metadata, staged_table.io, staged_table.metadata_location)
        tbl = self._convert_iceberg_into_hive(staged_table)

        with self._client as open_client:
            self._create_hive_table(open_client, tbl)
            hive_table = open_client.get_table(dbname=database_name, tbl_name=table_name)

        return self._convert_hive_into_iceberg(hive_table)

    def register_table(self, identifier: str | Identifier, metadata_location: str) -> Table:
        """Register a new table using existing metadata.

        Args:
            identifier (Union[str, Identifier]): Table identifier for the table
            metadata_location (str): The location to the metadata

        Returns:
            Table: The newly registered table

        Raises:
            TableAlreadyExistsError: If the table already exists
        """
        database_name, table_name = self.identifier_to_database_and_table(identifier)
        io = self._load_file_io(location=metadata_location)
        metadata_file = io.new_input(metadata_location)
        staged_table = StagedTable(
            identifier=(database_name, table_name),
            metadata=FromInputFile.table_metadata(metadata_file),
            metadata_location=metadata_location,
            io=io,
            catalog=self,
        )
        tbl = self._convert_iceberg_into_hive(staged_table)
        with self._client as open_client:
            self._create_hive_table(open_client, tbl)
            hive_table = open_client.get_table(dbname=database_name, tbl_name=table_name)

        return self._convert_hive_into_iceberg(hive_table)

    def list_views(self, namespace: str | Identifier) -> list[Identifier]:
        raise NotImplementedError

    def view_exists(self, identifier: str | Identifier) -> bool:
        raise NotImplementedError

    def _create_lock_request(self, database_name: str, table_name: str) -> LockRequest:
        lock_component: LockComponent = LockComponent(
            level=LockLevel.TABLE, type=LockType.EXCLUSIVE, dbname=database_name, tablename=table_name, isTransactional=True
        )

        lock_request: LockRequest = LockRequest(component=[lock_component], user=getpass.getuser(), hostname=socket.gethostname())

        return lock_request

    def _wait_for_lock(self, database_name: str, table_name: str, lockid: int, open_client: Client) -> LockResponse:
        @retry(
            retry=retry_if_exception_type(WaitingForLockException),
            wait=wait_exponential(multiplier=2, min=self._lock_check_min_wait_time, max=self._lock_check_max_wait_time),
            stop=stop_after_attempt(self._lock_check_retries),
            reraise=True,
        )
        def _do_wait_for_lock() -> LockResponse:
            response: LockResponse = open_client.check_lock(CheckLockRequest(lockid=lockid))
            if response.state == LockState.ACQUIRED:
                return response
            elif response.state == LockState.WAITING:
                msg = f"Wait on lock for {database_name}.{table_name}"
                logger.warning(msg)
                raise WaitingForLockException(msg)
            else:
                raise CommitFailedException(f"Failed to check lock for {database_name}.{table_name}, state: {response.state}")

        return _do_wait_for_lock()

    def commit_table(
        self, table: Table, requirements: tuple[TableRequirement, ...], updates: tuple[TableUpdate, ...]
    ) -> CommitTableResponse:
        """Commit updates to a table.

        Args:
            table (Table): The table to be updated.
            requirements: (Tuple[TableRequirement, ...]): Table requirements.
            updates: (Tuple[TableUpdate, ...]): Table updates.

        Returns:
            CommitTableResponse: The updated metadata.

        Raises:
            NoSuchTableError: If a table with the given identifier does not exist.
            CommitFailedException: Requirement not met, or a conflict with a concurrent commit.
        """
        table_identifier = table.name()
        database_name, table_name = self.identifier_to_database_and_table(table_identifier, NoSuchTableError)
        # commit to hive
        # https://github.com/apache/hive/blob/master/standalone-metastore/metastore-common/src/main/thrift/hive_metastore.thrift#L1232
        with self._client as open_client:
            lock: LockResponse = open_client.lock(self._create_lock_request(database_name, table_name))

            try:
                if lock.state != LockState.ACQUIRED:
                    if lock.state == LockState.WAITING:
                        self._wait_for_lock(database_name, table_name, lock.lockid, open_client)
                    else:
                        raise CommitFailedException(f"Failed to acquire lock for {table_identifier}, state: {lock.state}")

                hive_table: HiveTable | None
                current_table: Table | None
                try:
                    hive_table = self._get_hive_table(open_client, database_name, table_name)
                    current_table = self._convert_hive_into_iceberg(hive_table)
                except NoSuchTableError:
                    hive_table = None
                    current_table = None

                updated_staged_table = self._update_and_stage_table(current_table, table_identifier, requirements, updates)
                if current_table and updated_staged_table.metadata == current_table.metadata:
                    # no changes, do nothing
                    return CommitTableResponse(metadata=current_table.metadata, metadata_location=current_table.metadata_location)
                self._write_metadata(
                    metadata=updated_staged_table.metadata,
                    io=updated_staged_table.io,
                    metadata_path=updated_staged_table.metadata_location,
                )

                if hive_table and current_table:
                    # Table exists, update it.

                    # Note on table properties:
                    # - Iceberg table properties are stored in both HMS and Iceberg metadata JSON.
                    # - Updates are reflected in both locations
                    # - Existing HMS table properties (set by external systems like Hive/Spark) are preserved.
                    #
                    # While it is possible to modify HMS table properties through this API, it is not recommended:
                    # - Mixing HMS-specific properties in Iceberg metadata can cause confusion
                    # - New/updated HMS table properties will also be stored in Iceberg metadata (even though it is HMS-specific)
                    # - HMS-native properties (set outside Iceberg) cannot be deleted since they are not visible to Iceberg
                    #   (However, if you first SET an HMS property via Iceberg, it becomes tracked in Iceberg metadata,
                    #   and can then be deleted via Iceberg - which removes it from both Iceberg metadata and HMS)
                    new_iceberg_properties = _construct_parameters(
                        metadata_location=updated_staged_table.metadata_location,
                        previous_metadata_location=current_table.metadata_location,
                        metadata_properties=updated_staged_table.properties,
                    )
                    # Detect properties that were removed from Iceberg metadata
                    deleted_iceberg_properties = current_table.properties.keys() - updated_staged_table.properties.keys()

                    # Merge: preserve HMS-native properties, remove deleted Iceberg properties, apply new Iceberg properties
                    existing_hms_parameters = dict(hive_table.parameters or {})
                    for key in deleted_iceberg_properties:
                        existing_hms_parameters.pop(key, None)
                    existing_hms_parameters.update(new_iceberg_properties)
                    hive_table.parameters = existing_hms_parameters

                    # Update hive's schema and properties
                    hive_table.sd = _construct_hive_storage_descriptor(
                        updated_staged_table.schema(),
                        updated_staged_table.location(),
                        property_as_bool(self.properties, HIVE2_COMPATIBLE, HIVE2_COMPATIBLE_DEFAULT),
                    )
                    open_client.alter_table_with_environment_context(
                        dbname=database_name,
                        tbl_name=table_name,
                        new_tbl=hive_table,
                        environment_context=EnvironmentContext(properties={DO_NOT_UPDATE_STATS: DO_NOT_UPDATE_STATS_DEFAULT}),
                    )
                else:
                    # Table does not exist, create it.
                    hive_table = self._convert_iceberg_into_hive(
                        StagedTable(
                            identifier=(database_name, table_name),
                            metadata=updated_staged_table.metadata,
                            metadata_location=updated_staged_table.metadata_location,
                            io=updated_staged_table.io,
                            catalog=self,
                        )
                    )
                    self._create_hive_table(open_client, hive_table)
            except WaitingForLockException as e:
                raise CommitFailedException(f"Failed to acquire lock for {table_identifier}, state: {lock.state}") from e
            finally:
                open_client.unlock(UnlockRequest(lockid=lock.lockid))

        return CommitTableResponse(
            metadata=updated_staged_table.metadata, metadata_location=updated_staged_table.metadata_location
        )

    def load_table(self, identifier: str | Identifier) -> Table:
        """Load the table's metadata and return the table instance.

        You can also use this method to check for table existence using 'try catalog.table() except TableNotFoundError'.
        Note: This method doesn't scan data stored in the table.

        Args:
            identifier: Table identifier.

        Returns:
            Table: the table instance with its metadata.

        Raises:
            NoSuchTableError: If a table with the name does not exist, or the identifier is invalid.
        """
        database_name, table_name = self.identifier_to_database_and_table(identifier, NoSuchTableError)

        with self._client as open_client:
            hive_table = self._get_hive_table(open_client, database_name, table_name)

        return self._convert_hive_into_iceberg(hive_table)

    def drop_table(self, identifier: str | Identifier) -> None:
        """Drop a table.

        Args:
            identifier: Table identifier.

        Raises:
            NoSuchTableError: If a table with the name does not exist, or the identifier is invalid.
        """
        database_name, table_name = self.identifier_to_database_and_table(identifier, NoSuchTableError)
        try:
            with self._client as open_client:
                open_client.drop_table(dbname=database_name, name=table_name, deleteData=False)
        except NoSuchObjectException as e:
            # When the namespace doesn't exist, it throws the same error
            raise NoSuchTableError(f"Table does not exists: {table_name}") from e

    def purge_table(self, identifier: str | Identifier) -> None:
        # This requires to traverse the reachability set, and drop all the data files.
        raise NotImplementedError("Not yet implemented")

    def rename_table(self, from_identifier: str | Identifier, to_identifier: str | Identifier) -> Table:
        """Rename a fully classified table name.

        Args:
            from_identifier: Existing table identifier.
            to_identifier: New table identifier.

        Returns:
            Table: the updated table instance with its metadata.

        Raises:
            ValueError: When from table identifier is invalid.
            NoSuchTableError: When a table with the name does not exist.
            NoSuchNamespaceError: When the destination namespace doesn't exist.
            TableAlreadyExistsError: When the destination table already exists.
        """
        from_database_name, from_table_name = self.identifier_to_database_and_table(from_identifier, NoSuchTableError)
        to_database_name, to_table_name = self.identifier_to_database_and_table(to_identifier)

        if self.table_exists(to_identifier):
            raise TableAlreadyExistsError(f"Table already exists: {to_table_name}")

        try:
            with self._client as open_client:
                tbl = open_client.get_table(dbname=from_database_name, tbl_name=from_table_name)
                tbl.dbName = to_database_name
                tbl.tableName = to_table_name
                open_client.alter_table_with_environment_context(
                    dbname=from_database_name,
                    tbl_name=from_table_name,
                    new_tbl=tbl,
                    environment_context=EnvironmentContext(properties={DO_NOT_UPDATE_STATS: DO_NOT_UPDATE_STATS_DEFAULT}),
                )
        except NoSuchObjectException as e:
            raise NoSuchTableError(f"Table does not exist: {from_table_name}") from e
        except InvalidOperationException as e:
            raise NoSuchNamespaceError(f"Database does not exists: {to_database_name}") from e
        return self.load_table(to_identifier)

    def create_namespace(self, namespace: str | Identifier, properties: Properties = EMPTY_DICT) -> None:
        """Create a namespace in the catalog.

        Args:
            namespace: Namespace identifier.
            properties: A string dictionary of properties for the given namespace.

        Raises:
            ValueError: If the identifier is invalid.
            AlreadyExistsError: If a namespace with the given name already exists.
        """
        database_name = self.identifier_to_database(namespace)
        hive_database = HiveDatabase(name=database_name, parameters=properties)

        try:
            with self._client as open_client:
                open_client.create_database(_annotate_namespace(hive_database, properties))
        except AlreadyExistsException as e:
            raise NamespaceAlreadyExistsError(f"Database {database_name} already exists") from e

    def drop_namespace(self, namespace: str | Identifier) -> None:
        """Drop a namespace.

        Args:
            namespace: Namespace identifier.

        Raises:
            NoSuchNamespaceError: If a namespace with the given name does not exist, or the identifier is invalid.
            NamespaceNotEmptyError: If the namespace is not empty.
        """
        database_name = self.identifier_to_database(namespace, NoSuchNamespaceError)
        try:
            with self._client as open_client:
                open_client.drop_database(database_name, deleteData=False, cascade=False)
        except InvalidOperationException as e:
            raise NamespaceNotEmptyError(f"Database {database_name} is not empty") from e
        except MetaException as e:
            raise NoSuchNamespaceError(f"Database does not exists: {database_name}") from e

    def list_tables(self, namespace: str | Identifier) -> list[Identifier]:
        """List Iceberg tables under the given namespace in the catalog.

        When the database doesn't exist, it will just return an empty list.

        Args:
            namespace: Database to list.

        Returns:
            List[Identifier]: list of table identifiers.

        Raises:
            NoSuchNamespaceError: If a namespace with the given name does not exist, or the identifier is invalid.
        """
        database_name = self.identifier_to_database(namespace, NoSuchNamespaceError)
        with self._client as open_client:
            return [
                (database_name, table.tableName)
                for table in open_client.get_table_objects_by_name(
                    dbname=database_name, tbl_names=open_client.get_all_tables(db_name=database_name)
                )
                if table.parameters.get(TABLE_TYPE, "").lower() == ICEBERG
            ]

    def list_namespaces(self, namespace: str | Identifier = ()) -> list[Identifier]:
        """List namespaces from the given namespace. If not given, list top-level namespaces from the catalog.

        Returns:
            List[Identifier]: a List of namespace identifiers.
        """
        # Hierarchical namespace is not supported. Return an empty list
        if namespace:
            return []

        with self._client as open_client:
            return list(map(self.identifier_to_tuple, open_client.get_all_databases()))

    def load_namespace_properties(self, namespace: str | Identifier) -> Properties:
        """Get properties for a namespace.

        Args:
            namespace: Namespace identifier.

        Returns:
            Properties: Properties for the given namespace.

        Raises:
            NoSuchNamespaceError: If a namespace with the given name does not exist, or identifier is invalid.
        """
        database_name = self.identifier_to_database(namespace, NoSuchNamespaceError)
        try:
            with self._client as open_client:
                database = open_client.get_database(name=database_name)
                properties = database.parameters
                properties[LOCATION] = database.locationUri
                if comment := database.description:
                    properties[COMMENT] = comment
                return properties
        except NoSuchObjectException as e:
            raise NoSuchNamespaceError(f"Database does not exists: {database_name}") from e

    def update_namespace_properties(
        self, namespace: str | Identifier, removals: set[str] | None = None, updates: Properties = EMPTY_DICT
    ) -> PropertiesUpdateSummary:
        """Remove provided property keys and update properties for a namespace.

        Args:
            namespace: Namespace identifier.
            removals: Set of property keys that need to be removed. Optional Argument.
            updates: Properties to be updated for the given namespace.

        Raises:
            NoSuchNamespaceError: If a namespace with the given name does not exist
            ValueError: If removals and updates have overlapping keys.
        """
        self._check_for_overlap(updates=updates, removals=removals)
        database_name = self.identifier_to_database(namespace, NoSuchNamespaceError)
        with self._client as open_client:
            try:
                database = open_client.get_database(database_name)
                parameters = database.parameters
            except NoSuchObjectException as e:
                raise NoSuchNamespaceError(f"Database does not exists: {database_name}") from e

            removed: set[str] = set()
            updated: set[str] = set()

            if removals:
                for key in removals:
                    if key in parameters:
                        parameters.pop(key)
                        removed.add(key)
            if updates:
                for key, value in updates.items():
                    parameters[key] = value
                    updated.add(key)

            open_client.alter_database(database_name, _annotate_namespace(database, parameters))

        expected_to_change = (removals or set()).difference(removed)

        return PropertiesUpdateSummary(removed=list(removed or []), updated=list(updated or []), missing=list(expected_to_change))

    def drop_view(self, identifier: str | Identifier) -> None:
        raise NotImplementedError

    def _get_default_warehouse_location(self, database_name: str, table_name: str) -> str:
        """Override the default warehouse location to follow Hive-style conventions."""
        return self._get_hive_style_warehouse_location(database_name, table_name)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from unittest.mock import MagicMock
from uuid import uuid4

import pytest

from pyiceberg.table import CommitTableResponse, Table
from pyiceberg.table.update import SetSnapshotRefUpdate, TableUpdate


def _mock_commit_response(table: Table) -> CommitTableResponse:
    return CommitTableResponse(
        metadata=table.metadata,
        metadata_location="s3://bucket/tbl",
        uuid=uuid4(),
    )


def _get_updates(mock_catalog: MagicMock) -> tuple[TableUpdate, ...]:
    args, _ = mock_catalog.commit_table.call_args
    return args[2]


def test_set_current_snapshot_basic(table_v2: Table) -> None:
    snapshot_one = 3051729675574597004

    table_v2.catalog = MagicMock()
    table_v2.catalog.commit_table.return_value = _mock_commit_response(table_v2)

    table_v2.manage_snapshots().set_current_snapshot(snapshot_id=snapshot_one).commit()

    table_v2.catalog.commit_table.assert_called_once()

    updates = _get_updates(table_v2.catalog)
    set_ref_updates = [u for u in updates if isinstance(u, SetSnapshotRefUpdate)]

    assert len(set_ref_updates) == 1
    update = set_ref_updates[0]
    assert update.snapshot_id == snapshot_one
    assert update.ref_name == "main"
    assert update.type == "branch"


def test_set_current_snapshot_unknown_id(table_v2: Table) -> None:
    invalid_snapshot_id = 1234567890000
    table_v2.catalog = MagicMock()

    with pytest.raises(ValueError, match="Cannot set current snapshot to unknown snapshot id"):
        table_v2.manage_snapshots().set_current_snapshot(snapshot_id=invalid_snapshot_id).commit()

    table_v2.catalog.commit_table.assert_not_called()


def test_set_current_snapshot_to_current(table_v2: Table) -> None:
    current_snapshot = table_v2.current_snapshot()
    assert current_snapshot is not None

    table_v2.catalog = MagicMock()
    table_v2.catalog.commit_table.return_value = _mock_commit_response(table_v2)

    table_v2.manage_snapshots().set_current_snapshot(snapshot_id=current_snapshot.snapshot_id).commit()

    table_v2.catalog.commit_table.assert_called_once()


def test_set_current_snapshot_chained_with_tag(table_v2: Table) -> None:
    snapshot_one = 3051729675574597004
    table_v2.catalog = MagicMock()
    table_v2.catalog.commit_table.return_value = _mock_commit_response(table_v2)

    (table_v2.manage_snapshots().set_current_snapshot(snapshot_id=snapshot_one).create_tag(snapshot_one, "my-tag").commit())

    table_v2.catalog.commit_table.assert_called_once()

    updates = _get_updates(table_v2.catalog)
    set_ref_updates = [u for u in updates if isinstance(u, SetSnapshotRefUpdate)]

    assert len(set_ref_updates) == 2
    assert {u.ref_name for u in set_ref_updates} == {"main", "my-tag"}


def test_set_current_snapshot_with_extensive_snapshots(table_v2_with_extensive_snapshots: Table) -> None:
    snapshots = table_v2_with_extensive_snapshots.metadata.snapshots
    assert len(snapshots) > 100

    target_snapshot = snapshots[50].snapshot_id

    table_v2_with_extensive_snapshots.catalog = MagicMock()
    table_v2_with_extensive_snapshots.catalog.commit_table.return_value = _mock_commit_response(table_v2_with_extensive_snapshots)

    table_v2_with_extensive_snapshots.manage_snapshots().set_current_snapshot(snapshot_id=target_snapshot).commit()

    table_v2_with_extensive_snapshots.catalog.commit_table.assert_called_once()

    updates = _get_updates(table_v2_with_extensive_snapshots.catalog)
    set_ref_updates = [u for u in updates if isinstance(u, SetSnapshotRefUpdate)]

    assert len(set_ref_updates) == 1
    assert set_ref_updates[0].snapshot_id == target_snapshot


def test_set_current_snapshot_by_ref_name(table_v2: Table) -> None:
    current_snapshot = table_v2.current_snapshot()
    assert current_snapshot is not None

    table_v2.catalog = MagicMock()
    table_v2.catalog.commit_table.return_value = _mock_commit_response(table_v2)

    table_v2.manage_snapshots().set_current_snapshot(ref_name="main").commit()

    updates = _get_updates(table_v2.catalog)
    set_ref_updates = [u for u in updates if isinstance(u, SetSnapshotRefUpdate)]

    assert len(set_ref_updates) == 1
    assert set_ref_updates[0].snapshot_id == current_snapshot.snapshot_id
    assert set_ref_updates[0].ref_name == "main"


def test_set_current_snapshot_unknown_ref(table_v2: Table) -> None:
    table_v2.catalog = MagicMock()

    with pytest.raises(ValueError, match="Cannot find matching snapshot ID for ref: nonexistent"):
        table_v2.manage_snapshots().set_current_snapshot(ref_name="nonexistent").commit()

    table_v2.catalog.commit_table.assert_not_called()


def test_set_current_snapshot_requires_one_argument(table_v2: Table) -> None:
    table_v2.catalog = MagicMock()

    with pytest.raises(ValueError, match="Either snapshot_id or ref_name must be provided, not both"):
        table_v2.manage_snapshots().set_current_snapshot().commit()

    with pytest.raises(ValueError, match="Either snapshot_id or ref_name must be provided, not both"):
        table_v2.manage_snapshots().set_current_snapshot(snapshot_id=123, ref_name="main").commit()

    table_v2.catalog.commit_table.assert_not_called()


def test_set_current_snapshot_chained_with_create_tag(table_v2: Table) -> None:
    snapshot_one = 3051729675574597004
    table_v2.catalog = MagicMock()
    table_v2.catalog.commit_table.return_value = _mock_commit_response(table_v2)

    # create a tag and immediately use it to set current snapshot
    (
        table_v2.manage_snapshots()
        .create_tag(snapshot_id=snapshot_one, tag_name="new-tag")
        .set_current_snapshot(ref_name="new-tag")
        .commit()
    )

    table_v2.catalog.commit_table.assert_called_once()

    updates = _get_updates(table_v2.catalog)
    set_ref_updates = [u for u in updates if isinstance(u, SetSnapshotRefUpdate)]

    # should have the tag and the main branch update
    assert len(set_ref_updates) == 2
    assert {u.ref_name for u in set_ref_updates} == {"new-tag", "main"}

    # The main branch should point to the same snapshot as the tag
    main_update = next(u for u in set_ref_updates if u.ref_name == "main")
    assert main_update.snapshot_id == snapshot_one



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence
from functools import cached_property
from typing import Any, TypeAlias
from typing import Literal as TypingLiteral

from pydantic import ConfigDict, Field, SerializeAsAny, model_validator
from pydantic_core.core_schema import ValidatorFunctionWrapHandler

from pyiceberg.expressions.literals import AboveMax, BelowMin, Literal, literal
from pyiceberg.schema import Accessor, Schema
from pyiceberg.typedef import IcebergBaseModel, IcebergRootModel, L, LiteralValue, StructProtocol
from pyiceberg.types import DoubleType, FloatType, NestedField
from pyiceberg.utils.singleton import Singleton


def _to_unbound_term(term: str | UnboundTerm) -> UnboundTerm:
    return Reference(term) if isinstance(term, str) else term


def _to_literal_set(values: Iterable[L] | Iterable[Literal[L]]) -> set[Literal[L]]:
    return {_to_literal(v) for v in values}


def _to_literal(value: L | Literal[L]) -> Literal[L]:
    if isinstance(value, Literal):
        return value
    else:
        return literal(value)


class BooleanExpression(IcebergBaseModel, ABC):
    """An expression that evaluates to a boolean."""

    @abstractmethod
    def __invert__(self) -> BooleanExpression:
        """Transform the Expression into its negated version."""

    def __and__(self, other: BooleanExpression) -> BooleanExpression:
        """Perform and operation on another expression."""
        if not isinstance(other, BooleanExpression):
            raise ValueError(f"Expected BooleanExpression, got: {other}")

        return And(self, other)

    def __or__(self, other: BooleanExpression) -> BooleanExpression:
        """Perform or operation on another expression."""
        if not isinstance(other, BooleanExpression):
            raise ValueError(f"Expected BooleanExpression, got: {other}")

        return Or(self, other)

    @model_validator(mode="wrap")
    @classmethod
    def handle_primitive_type(cls, v: Any, handler: ValidatorFunctionWrapHandler) -> BooleanExpression:
        """Apply custom deserialization logic before validation."""
        # Already a BooleanExpression? return as-is so we keep the concrete subclass.
        if isinstance(v, BooleanExpression):
            return v

        # Handle different input formats
        if isinstance(v, bool):
            return AlwaysTrue() if v is True else AlwaysFalse()

        if isinstance(v, dict) and (field_type := v.get("type")):
            # Unary
            if field_type == "is-null":
                return IsNull(**v)
            elif field_type == "not-null":
                return NotNull(**v)
            elif field_type == "is-nan":
                return IsNaN(**v)
            elif field_type == "not-nan":
                return NotNaN(**v)

            # Literal
            elif field_type == "lt":
                return LessThan(**v)
            elif field_type == "lt-eq":
                return LessThanOrEqual(**v)
            elif field_type == "gt":
                return GreaterThan(**v)
            elif field_type == "gt-eq":
                return GreaterThanOrEqual(**v)
            elif field_type == "eq":
                return EqualTo(**v)
            elif field_type == "not-eq":
                return NotEqualTo(**v)
            elif field_type == "starts-with":
                return StartsWith(**v)
            elif field_type == "not-starts-with":
                return NotStartsWith(**v)

            # Set
            elif field_type == "in":
                return In(**v)
            elif field_type == "not-in":
                return NotIn(**v)

            # Other
            elif field_type == "and":
                return And(**v)
            elif field_type == "or":
                return Or(**v)
            elif field_type == "not":
                return Not(**v)

        return handler(v)


SerializableBooleanExpression: TypeAlias = SerializeAsAny["BooleanExpression"]


def _build_balanced_tree(
    operator_: Callable[[BooleanExpression, BooleanExpression], BooleanExpression], items: Sequence[BooleanExpression]
) -> BooleanExpression:
    """
    Recursively constructs a balanced binary tree of BooleanExpressions using the provided binary operator.

    This function is a safer and more scalable alternative to:
        reduce(operator_, items)

    Using `reduce` creates a deeply nested, unbalanced tree (e.g., operator_(a, operator_(b, operator_(c, ...)))),
    which grows linearly with the number of items. This can lead to RecursionError exceptions in Python
    when the number of expressions is large (e.g., >1000).

    In contrast, this function builds a balanced binary tree with logarithmic depth (O(log n)),
    helping avoid recursion issues and ensuring that expression trees remain stable, predictable,
    and safe to traverse — especially in tools like PyIceberg that operate on large logical trees.

    Parameters:
        operator_ (Callable): A binary operator function (e.g., pyiceberg.expressions.Or, And) that takes two
            BooleanExpressions and returns a combined BooleanExpression.
        items (Sequence[BooleanExpression]): A sequence of BooleanExpression objects to combine.

    Returns:
        BooleanExpression: The balanced combination of all input BooleanExpressions.

    Raises:
        ValueError: If the input sequence is empty.
    """
    if not items:
        raise ValueError("No expressions to combine")
    if len(items) == 1:
        return items[0]
    mid = len(items) // 2

    left = _build_balanced_tree(operator_, items[:mid])
    right = _build_balanced_tree(operator_, items[mid:])
    return operator_(left, right)


class Term:
    """A simple expression that evaluates to a value."""


class Bound:
    """Represents a bound value expression."""


class Unbound(ABC):
    """Represents an unbound value expression."""

    @abstractmethod
    def bind(self, schema: Schema, case_sensitive: bool = True) -> Bound | BooleanExpression: ...

    @property
    @abstractmethod
    def as_bound(self) -> type[Bound]: ...


class BoundTerm(Term, Bound, ABC):
    """Represents a bound term."""

    @abstractmethod
    def ref(self) -> BoundReference:
        """Return the bound reference."""

    @abstractmethod
    def eval(self, struct: StructProtocol) -> Any:  # pylint: disable=W0613
        """Return the value at the referenced field's position in an object that abides by the StructProtocol."""


class BoundReference(BoundTerm):
    """A reference bound to a field in a schema.

    Args:
        field (NestedField): A referenced field in an Iceberg schema.
        accessor (Accessor): An Accessor object to access the value at the field's position.
    """

    field: NestedField
    accessor: Accessor

    def __init__(self, field: NestedField, accessor: Accessor):
        self.field = field
        self.accessor = accessor

    def eval(self, struct: StructProtocol) -> Any:
        """Return the value at the referenced field's position in an object that abides by the StructProtocol.

        Args:
            struct (StructProtocol): A row object that abides by the StructProtocol and returns values given a position.
        Returns:
            Any: The value at the referenced field's position in `struct`.
        """
        return self.accessor.get(struct)

    def __eq__(self, other: Any) -> bool:
        """Return the equality of two instances of the BoundReference class."""
        return self.field == other.field if isinstance(other, BoundReference) else False

    def __repr__(self) -> str:
        """Return the string representation of the BoundReference class."""
        return f"BoundReference(field={repr(self.field)}, accessor={repr(self.accessor)})"

    def ref(self) -> BoundReference:
        return self

    def __hash__(self) -> int:
        """Return hash value of the BoundReference class."""
        return hash(str(self))


class UnboundTerm(Term, Unbound, ABC):
    """Represents an unbound term."""

    @abstractmethod
    def bind(self, schema: Schema, case_sensitive: bool = True) -> BoundTerm: ...


class Reference(UnboundTerm, IcebergRootModel[str]):
    """A reference not yet bound to a field in a schema.

    Args:
        name (str): The name of the field.

    Note:
        An unbound reference is sometimes referred to as a "named" reference.
    """

    root: str = Field()

    def __init__(self, name: str) -> None:
        super().__init__(name)

    def __repr__(self) -> str:
        """Return the string representation of the Reference class."""
        return f"Reference(name={repr(self.root)})"

    def __str__(self) -> str:
        """Return the string representation of the Reference class."""
        return f"Reference(name={repr(self.root)})"

    def bind(self, schema: Schema, case_sensitive: bool = True) -> BoundReference:
        """Bind the reference to an Iceberg schema.

        Args:
            schema (Schema): An Iceberg schema.
            case_sensitive (bool): Whether to consider case when binding the reference to the field.

        Raises:
            ValueError: If an empty name is provided.

        Returns:
            BoundReference: A reference bound to the specific field in the Iceberg schema.
        """
        field = schema.find_field(name_or_id=self.name, case_sensitive=case_sensitive)
        accessor = schema.accessor_for_field(field.field_id)
        return self.as_bound(field=field, accessor=accessor)

    @property
    def name(self) -> str:
        return self.root

    @property
    def as_bound(self) -> type[BoundReference]:
        return BoundReference


class And(BooleanExpression):
    """AND operation expression - logical conjunction."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    type: TypingLiteral["and"] = Field(default="and", alias="type")
    left: SerializableBooleanExpression = Field()
    right: SerializableBooleanExpression = Field()

    def __init__(self, left: BooleanExpression, right: BooleanExpression, *rest: BooleanExpression, **_: Any) -> None:
        if isinstance(self, And) and not hasattr(self, "left") and not hasattr(self, "right"):
            super().__init__(left=left, right=right)

    def __new__(cls, left: BooleanExpression, right: BooleanExpression, *rest: BooleanExpression, **_: Any) -> BooleanExpression:
        if rest:
            return _build_balanced_tree(And, (left, right, *rest))
        if left is AlwaysFalse() or right is AlwaysFalse():
            return AlwaysFalse()
        elif left is AlwaysTrue():
            return right
        elif right is AlwaysTrue():
            return left
        else:
            return super().__new__(cls)

    def __eq__(self, other: Any) -> bool:
        """Return the equality of two instances of the And class."""
        return self.left == other.left and self.right == other.right if isinstance(other, And) else False

    def __str__(self) -> str:
        """Return the string representation of the And class."""
        return f"And(left={str(self.left)}, right={str(self.right)})"

    def __repr__(self) -> str:
        """Return the string representation of the And class."""
        return f"And(left={repr(self.left)}, right={repr(self.right)})"

    def __invert__(self) -> BooleanExpression:
        """Transform the Expression into its negated version."""
        # De Morgan's law: not (A and B) = (not A) or (not B)
        return Or(~self.left, ~self.right)

    def __getnewargs__(self) -> tuple[BooleanExpression, BooleanExpression]:
        """Pickle the And class."""
        return (self.left, self.right)


class Or(BooleanExpression):
    """OR operation expression - logical disjunction."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    type: TypingLiteral["or"] = Field(default="or", alias="type")
    left: SerializableBooleanExpression = Field()
    right: SerializableBooleanExpression = Field()

    def __init__(self, left: BooleanExpression, right: BooleanExpression, *rest: BooleanExpression, **_: Any) -> None:
        if isinstance(self, Or) and not hasattr(self, "left") and not hasattr(self, "right"):
            super().__init__(left=left, right=right)

    def __new__(cls, left: BooleanExpression, right: BooleanExpression, *rest: BooleanExpression, **_: Any) -> BooleanExpression:
        if rest:
            return _build_balanced_tree(Or, (left, right, *rest))
        if left is AlwaysTrue() or right is AlwaysTrue():
            return AlwaysTrue()
        elif left is AlwaysFalse():
            return right
        elif right is AlwaysFalse():
            return left
        else:
            return super().__new__(cls)

    def __str__(self) -> str:
        """Return the string representation of the Or class."""
        return f"{str(self.__class__.__name__)}(left={repr(self.left)}, right={repr(self.right)})"

    def __eq__(self, other: Any) -> bool:
        """Return the equality of two instances of the Or class."""
        return self.left == other.left and self.right == other.right if isinstance(other, Or) else False

    def __repr__(self) -> str:
        """Return the string representation of the Or class."""
        return f"Or(left={repr(self.left)}, right={repr(self.right)})"

    def __invert__(self) -> BooleanExpression:
        """Transform the Expression into its negated version."""
        # De Morgan's law: not (A or B) = (not A) and (not B)
        return And(~self.left, ~self.right)

    def __getnewargs__(self) -> tuple[BooleanExpression, BooleanExpression]:
        """Pickle the Or class."""
        return (self.left, self.right)


class Not(BooleanExpression):
    """NOT operation expression - logical negation."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    type: TypingLiteral["not"] = Field(default="not")
    child: SerializableBooleanExpression = Field()

    def __init__(self, child: BooleanExpression, **_: Any) -> None:
        super().__init__(child=child)

    def __new__(cls, child: BooleanExpression, **_: Any) -> BooleanExpression:
        if child is AlwaysTrue():
            return AlwaysFalse()
        elif child is AlwaysFalse():
            return AlwaysTrue()
        elif isinstance(child, Not):
            return child.child
        else:
            return super().__new__(cls)

    def __str__(self) -> str:
        """Return the string representation of the Not class."""
        return f"Not(child={self.child})"

    def __repr__(self) -> str:
        """Return the string representation of the Not class."""
        return f"Not(child={repr(self.child)})"

    def __eq__(self, other: Any) -> bool:
        """Return the equality of two instances of the Not class."""
        return self.child == other.child if isinstance(other, Not) else False

    def __invert__(self) -> BooleanExpression:
        """Transform the Expression into its negated version."""
        return self.child

    def __getnewargs__(self) -> tuple[BooleanExpression]:
        """Pickle the Not class."""
        return (self.child,)


class AlwaysTrue(BooleanExpression, Singleton, IcebergRootModel[bool]):
    """TRUE expression."""

    root: bool = True

    def __invert__(self) -> AlwaysFalse:
        """Transform the Expression into its negated version."""
        return AlwaysFalse()

    def __str__(self) -> str:
        """Return the string representation of the AlwaysTrue class."""
        return "AlwaysTrue()"

    def __repr__(self) -> str:
        """Return the string representation of the AlwaysTrue class."""
        return "AlwaysTrue()"


class AlwaysFalse(BooleanExpression, Singleton, IcebergRootModel[bool]):
    """FALSE expression."""

    root: bool = False

    def __invert__(self) -> AlwaysTrue:
        """Transform the Expression into its negated version."""
        return AlwaysTrue()

    def __str__(self) -> str:
        """Return the string representation of the AlwaysFalse class."""
        return "AlwaysFalse()"

    def __repr__(self) -> str:
        """Return the string representation of the AlwaysFalse class."""
        return "AlwaysFalse()"


class BoundPredicate(Bound, BooleanExpression, ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    term: BoundTerm

    def __init__(self, term: BoundTerm, **kwargs: Any) -> None:
        super().__init__(term=term, **kwargs)

    def __eq__(self, other: Any) -> bool:
        """Return the equality of two instances of the BoundPredicate class."""
        if isinstance(other, self.__class__):
            return self.term == other.term
        return False

    def __str__(self) -> str:
        """Return the string representation of the BoundPredicate class."""
        return f"{self.__class__.__name__}(term={str(self.term)})"

    @property
    @abstractmethod
    def as_unbound(self) -> type[UnboundPredicate]: ...


class UnboundPredicate(Unbound, BooleanExpression, ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    term: UnboundTerm

    def __init__(self, term: str | UnboundTerm, **kwargs: Any) -> None:
        super().__init__(term=_to_unbound_term(term), **kwargs)

    def __eq__(self, other: Any) -> bool:
        """Return the equality of two instances of the UnboundPredicate class."""
        return self.term == other.term if isinstance(other, self.__class__) else False

    @abstractmethod
    def bind(self, schema: Schema, case_sensitive: bool = True) -> BooleanExpression: ...

    @property
    @abstractmethod
    def as_bound(self) -> type[BoundPredicate]: ...


class UnaryPredicate(UnboundPredicate, ABC):
    type: TypingLiteral["is-null", "not-null", "is-nan", "not-nan"] = Field()

    model_config = {"arbitrary_types_allowed": True}

    def __init__(self, term: str | UnboundTerm, **_: Any) -> None:
        unbound = _to_unbound_term(term)
        super().__init__(term=unbound)

    def __str__(self) -> str:
        """Return the string representation of the UnaryPredicate class."""
        # Sort to make it deterministic
        return f"{str(self.__class__.__name__)}(term={str(self.term)})"

    def bind(self, schema: Schema, case_sensitive: bool = True) -> BoundUnaryPredicate:
        bound_term = self.term.bind(schema, case_sensitive)
        bound_type = self.as_bound
        return bound_type(bound_term)  # type: ignore[misc]

    def __repr__(self) -> str:
        """Return the string representation of the UnaryPredicate class."""
        return f"{str(self.__class__.__name__)}(term={repr(self.term)})"

    @property
    @abstractmethod
    def as_bound(self) -> type[BoundUnaryPredicate]: ...  # type: ignore


class BoundUnaryPredicate(BoundPredicate, ABC):
    def __repr__(self) -> str:
        """Return the string representation of the BoundUnaryPredicate class."""
        return f"{str(self.__class__.__name__)}(term={repr(self.term)})"

    @property
    @abstractmethod
    def as_unbound(self) -> type[UnaryPredicate]: ...

    def __getnewargs__(self) -> tuple[BoundTerm]:
        """Pickle the BoundUnaryPredicate class."""
        return (self.term,)


class BoundIsNull(BoundUnaryPredicate):
    def __new__(cls, term: BoundTerm) -> BooleanExpression:  # pylint: disable=W0221
        if term.ref().field.required:
            return AlwaysFalse()
        return super().__new__(cls)

    def __invert__(self) -> BoundNotNull:
        """Transform the Expression into its negated version."""
        return BoundNotNull(self.term)

    @property
    def as_unbound(self) -> type[IsNull]:
        return IsNull


class BoundNotNull(BoundUnaryPredicate):
    def __new__(cls, term: BoundTerm) -> BooleanExpression:  # pylint: disable=W0221
        if term.ref().field.required:
            return AlwaysTrue()
        return super().__new__(cls)

    def __invert__(self) -> BoundIsNull:
        """Transform the Expression into its negated version."""
        return BoundIsNull(self.term)

    @property
    def as_unbound(self) -> type[NotNull]:
        return NotNull


class IsNull(UnaryPredicate):
    type: TypingLiteral["is-null"] = Field(default="is-null")

    def __invert__(self) -> NotNull:
        """Transform the Expression into its negated version."""
        return NotNull(self.term)

    @property
    def as_bound(self) -> type[BoundIsNull]:  # type: ignore
        return BoundIsNull


class NotNull(UnaryPredicate):
    type: TypingLiteral["not-null"] = Field(default="not-null")

    def __invert__(self) -> IsNull:
        """Transform the Expression into its negated version."""
        return IsNull(self.term)

    @property
    def as_bound(self) -> type[BoundNotNull]:  # type: ignore
        return BoundNotNull


class BoundIsNaN(BoundUnaryPredicate):
    def __new__(cls, term: BoundTerm) -> BooleanExpression:  # pylint: disable=W0221
        bound_type = term.ref().field.field_type
        if isinstance(bound_type, (FloatType, DoubleType)):
            return super().__new__(cls)
        return AlwaysFalse()

    def __invert__(self) -> BoundNotNaN:
        """Transform the Expression into its negated version."""
        return BoundNotNaN(self.term)

    @property
    def as_unbound(self) -> type[IsNaN]:
        return IsNaN


class BoundNotNaN(BoundUnaryPredicate):
    def __new__(cls, term: BoundTerm) -> BooleanExpression:  # pylint: disable=W0221
        bound_type = term.ref().field.field_type
        if isinstance(bound_type, (FloatType, DoubleType)):
            return super().__new__(cls)
        return AlwaysTrue()

    def __invert__(self) -> BoundIsNaN:
        """Transform the Expression into its negated version."""
        return BoundIsNaN(self.term)

    @property
    def as_unbound(self) -> type[NotNaN]:
        return NotNaN


class IsNaN(UnaryPredicate):
    type: TypingLiteral["is-nan"] = Field(default="is-nan")

    def __invert__(self) -> NotNaN:
        """Transform the Expression into its negated version."""
        return NotNaN(self.term)

    @property
    def as_bound(self) -> type[BoundIsNaN]:  # type: ignore
        return BoundIsNaN


class NotNaN(UnaryPredicate):
    type: TypingLiteral["not-nan"] = Field(default="not-nan")

    def __invert__(self) -> IsNaN:
        """Transform the Expression into its negated version."""
        return IsNaN(self.term)

    @property
    def as_bound(self) -> type[BoundNotNaN]:  # type: ignore
        return BoundNotNaN


class SetPredicate(UnboundPredicate, ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    type: TypingLiteral["in", "not-in"] = Field(default="in")
    literals: set[LiteralValue] = Field(alias="values")

    def __init__(
        self, term: str | UnboundTerm, literals: Iterable[Any] | Iterable[LiteralValue] | None = None, **kwargs: Any
    ) -> None:
        if literals is None and "values" in kwargs:
            literals = kwargs["values"]

        if literals is None:
            literal_set: set[LiteralValue] = set()
        else:
            literal_set = _to_literal_set(literals)
        super().__init__(term=_to_unbound_term(term), values=literal_set)

    def bind(self, schema: Schema, case_sensitive: bool = True) -> BoundSetPredicate:
        bound_term = self.term.bind(schema, case_sensitive)
        literal_set = self.literals
        return self.as_bound(bound_term, {lit.to(bound_term.ref().field.field_type) for lit in literal_set})  # type: ignore

    def __str__(self) -> str:
        """Return the string representation of the SetPredicate class."""
        # Sort to make it deterministic
        literals_str = ", ".join(sorted([str(literal) for literal in self.literals]))
        return f"{str(self.__class__.__name__)}({str(self.term)}, {{{literals_str}}})"

    def __repr__(self) -> str:
        """Return the string representation of the SetPredicate class."""
        # Sort to make it deterministic
        literals_repr = ", ".join(sorted([repr(literal) for literal in self.literals]))
        return f"{str(self.__class__.__name__)}({repr(self.term)}, {{{literals_repr}}})"

    def __eq__(self, other: Any) -> bool:
        """Return the equality of two instances of the SetPredicate class."""
        return self.term == other.term and self.literals == other.literals if isinstance(other, self.__class__) else False

    def __getnewargs__(self) -> tuple[UnboundTerm, set[Any]]:
        """Pickle the SetPredicate class."""
        return (self.term, self.literals)

    @property
    @abstractmethod
    def as_bound(self) -> type[BoundSetPredicate]:  # type: ignore
        return BoundSetPredicate


class BoundSetPredicate(BoundPredicate, ABC):
    literals: set[LiteralValue]

    def __init__(self, term: BoundTerm, literals: set[LiteralValue]) -> None:
        literal_set = _to_literal_set(literals)
        super().__init__(term=term, literals=literal_set)

    @cached_property
    def value_set(self) -> set[Any]:
        return {lit.value for lit in self.literals}

    def __str__(self) -> str:
        """Return the string representation of the BoundSetPredicate class."""
        # Sort to make it deterministic
        literals_str = ", ".join(sorted([str(literal) for literal in self.literals]))
        return f"{str(self.__class__.__name__)}({str(self.term)}, {{{literals_str}}})"

    def __repr__(self) -> str:
        """Return the string representation of the BoundSetPredicate class."""
        # Sort to make it deterministic
        literals_repr = ", ".join(sorted([repr(literal) for literal in self.literals]))
        return f"{str(self.__class__.__name__)}({repr(self.term)}, {{{literals_repr}}})"

    def __eq__(self, other: Any) -> bool:
        """Return the equality of two instances of the BoundSetPredicate class."""
        return self.term == other.term and self.literals == other.literals if isinstance(other, self.__class__) else False

    def __getnewargs__(self) -> tuple[BoundTerm, set[LiteralValue]]:
        """Pickle the BoundSetPredicate class."""
        return (self.term, self.literals)

    @property
    @abstractmethod
    def as_unbound(self) -> type[SetPredicate]: ...


class BoundIn(BoundSetPredicate):
    def __new__(cls, term: BoundTerm, literals: set[LiteralValue]) -> BooleanExpression:  # pylint: disable=W0221
        count = len(literals)
        if count == 0:
            return AlwaysFalse()
        elif count == 1:
            return BoundEqualTo(term, next(iter(literals)))
        else:
            return super().__new__(cls)

    def __invert__(self) -> BoundNotIn:
        """Transform the Expression into its negated version."""
        return BoundNotIn(self.term, self.literals)

    def __eq__(self, other: Any) -> bool:
        """Return the equality of two instances of the BoundIn class."""
        return self.term == other.term and self.literals == other.literals if isinstance(other, self.__class__) else False

    @property
    def as_unbound(self) -> type[In]:
        return In


class BoundNotIn(BoundSetPredicate):
    def __new__(  # pylint: disable=W0221
        cls,
        term: BoundTerm,
        literals: set[LiteralValue],
    ) -> BooleanExpression:
        count = len(literals)
        if count == 0:
            return AlwaysTrue()
        elif count == 1:
            return BoundNotEqualTo(term, next(iter(literals)))
        else:
            return super().__new__(cls)

    def __invert__(self) -> BoundIn:
        """Transform the Expression into its negated version."""
        return BoundIn(self.term, self.literals)

    @property
    def as_unbound(self) -> type[NotIn]:
        return NotIn


class In(SetPredicate):
    type: TypingLiteral["in"] = Field(default="in", alias="type")

    def __new__(  # pylint: disable=W0221
        cls, term: str | UnboundTerm, literals: Iterable[Any] | Iterable[LiteralValue] | None = None, **kwargs: Any
    ) -> In:
        if literals is None and "values" in kwargs:
            literals = kwargs["values"]

        if literals is None:
            literals_set: set[LiteralValue] = set()
        else:
            literals_set = _to_literal_set(literals)
        count = len(literals_set)
        if count == 0:
            return AlwaysFalse()
        elif count == 1:
            return EqualTo(term, next(iter(literals_set)))
        else:
            return super().__new__(cls)

    def __invert__(self) -> NotIn:
        """Transform the Expression into its negated version."""
        return NotIn(self.term, self.literals)

    @property
    def as_bound(self) -> type[BoundIn]:  # type: ignore
        return BoundIn


class NotIn(SetPredicate, ABC):
    type: TypingLiteral["not-in"] = Field(default="not-in", alias="type")

    def __new__(  # pylint: disable=W0221
        cls, term: str | UnboundTerm, literals: Iterable[Any] | Iterable[LiteralValue] | None = None, **kwargs: Any
    ) -> NotIn:
        if literals is None and "values" in kwargs:
            literals = kwargs["values"]

        if literals is None:
            literals_set: set[LiteralValue] = set()
        else:
            literals_set = _to_literal_set(literals)
        count = len(literals_set)
        if count == 0:
            return AlwaysTrue()
        elif count == 1:
            return NotEqualTo(term, next(iter(literals_set)))
        else:
            return super().__new__(cls)

    def __invert__(self) -> In:
        """Transform the Expression into its negated version."""
        return In(self.term, self.literals)

    @property
    def as_bound(self) -> type[BoundNotIn]:  # type: ignore
        return BoundNotIn


class LiteralPredicate(UnboundPredicate, ABC):
    type: TypingLiteral["lt", "lt-eq", "gt", "gt-eq", "eq", "not-eq", "starts-with", "not-starts-with"] = Field(alias="type")
    term: UnboundTerm
    value: LiteralValue = Field()
    model_config = ConfigDict(populate_by_name=True, frozen=True, arbitrary_types_allowed=True)

    def __init__(self, term: str | UnboundTerm, literal: Any | None = None, **kwargs: Any) -> None:
        if literal is None and "value" in kwargs:
            literal = kwargs["value"]

        super().__init__(term=_to_unbound_term(term), value=_to_literal(literal))

    @property
    def literal(self) -> LiteralValue:
        return self.value

    def bind(self, schema: Schema, case_sensitive: bool = True) -> BoundLiteralPredicate:
        bound_term = self.term.bind(schema, case_sensitive)
        lit = self.literal.to(bound_term.ref().field.field_type)

        if isinstance(lit, AboveMax):
            if isinstance(self, (LessThan, LessThanOrEqual, NotEqualTo)):
                return AlwaysTrue()
            elif isinstance(self, (GreaterThan, GreaterThanOrEqual, EqualTo)):
                return AlwaysFalse()
        elif isinstance(lit, BelowMin):
            if isinstance(self, (GreaterThan, GreaterThanOrEqual, NotEqualTo)):
                return AlwaysTrue()
            elif isinstance(self, (LessThan, LessThanOrEqual, EqualTo)):
                return AlwaysFalse()

        return self.as_bound(bound_term, lit)  # type: ignore

    def __eq__(self, other: Any) -> bool:
        """Return the equality of two instances of the LiteralPredicate class."""
        if isinstance(other, self.__class__):
            return self.term == other.term and self.literal == other.literal
        return False

    def __str__(self) -> str:
        """Return the string representation of the LiteralPredicate class."""
        return f"{str(self.__class__.__name__)}(term={repr(self.term)}, literal={repr(self.literal)})"

    def __repr__(self) -> str:
        """Return the string representation of the LiteralPredicate class."""
        return f"{str(self.__class__.__name__)}(term={repr(self.term)}, literal={repr(self.literal)})"

    @property
    @abstractmethod
    def as_bound(self) -> type[BoundLiteralPredicate]: ...  # type: ignore


class BoundLiteralPredicate(BoundPredicate, ABC):
    literal: LiteralValue

    def __init__(self, term: BoundTerm, literal: LiteralValue):  # pylint: disable=W0621
        super().__init__(term=term, literal=literal)

    def __eq__(self, other: Any) -> bool:
        """Return the equality of two instances of the BoundLiteralPredicate class."""
        if isinstance(other, self.__class__):
            return self.term == other.term and self.literal == other.literal
        return False

    def __str__(self) -> str:
        """Return the string representation of the BoundLiteralPredicate class."""
        return f"{self.__class__.__name__}(term={str(self.term)}, literal={repr(self.literal)})"

    def __repr__(self) -> str:
        """Return the string representation of the BoundLiteralPredicate class."""
        return f"{str(self.__class__.__name__)}(term={repr(self.term)}, literal={repr(self.literal)})"

    @property
    @abstractmethod
    def as_unbound(self) -> type[LiteralPredicate]: ...


class BoundEqualTo(BoundLiteralPredicate):
    def __invert__(self) -> BoundNotEqualTo:
        """Transform the Expression into its negated version."""
        return BoundNotEqualTo(self.term, self.literal)

    @property
    def as_unbound(self) -> type[EqualTo]:
        return EqualTo


class BoundNotEqualTo(BoundLiteralPredicate):
    def __invert__(self) -> BoundEqualTo:
        """Transform the Expression into its negated version."""
        return BoundEqualTo(self.term, self.literal)

    @property
    def as_unbound(self) -> type[NotEqualTo]:
        return NotEqualTo


class BoundGreaterThanOrEqual(BoundLiteralPredicate):
    def __invert__(self) -> BoundLessThan:
        """Transform the Expression into its negated version."""
        return BoundLessThan(self.term, self.literal)

    @property
    def as_unbound(self) -> type[GreaterThanOrEqual]:
        return GreaterThanOrEqual


class BoundGreaterThan(BoundLiteralPredicate):
    def __invert__(self) -> BoundLessThanOrEqual:
        """Transform the Expression into its negated version."""
        return BoundLessThanOrEqual(self.term, self.literal)

    @property
    def as_unbound(self) -> type[GreaterThan]:
        return GreaterThan


class BoundLessThan(BoundLiteralPredicate):
    def __invert__(self) -> BoundGreaterThanOrEqual:
        """Transform the Expression into its negated version."""
        return BoundGreaterThanOrEqual(self.term, self.literal)

    @property
    def as_unbound(self) -> type[LessThan]:
        return LessThan


class BoundLessThanOrEqual(BoundLiteralPredicate):
    def __invert__(self) -> BoundGreaterThan:
        """Transform the Expression into its negated version."""
        return BoundGreaterThan(self.term, self.literal)

    @property
    def as_unbound(self) -> type[LessThanOrEqual]:
        return LessThanOrEqual


class BoundStartsWith(BoundLiteralPredicate):
    def __invert__(self) -> BoundNotStartsWith:
        """Transform the Expression into its negated version."""
        return BoundNotStartsWith(self.term, self.literal)

    @property
    def as_unbound(self) -> type[StartsWith]:
        return StartsWith


class BoundNotStartsWith(BoundLiteralPredicate):
    def __invert__(self) -> BoundStartsWith:
        """Transform the Expression into its negated version."""
        return BoundStartsWith(self.term, self.literal)

    @property
    def as_unbound(self) -> type[NotStartsWith]:
        return NotStartsWith


class EqualTo(LiteralPredicate):
    type: TypingLiteral["eq"] = Field(default="eq", alias="type")

    def __invert__(self) -> NotEqualTo:
        """Transform the Expression into its negated version."""
        return NotEqualTo(self.term, self.literal)

    @property
    def as_bound(self) -> type[BoundEqualTo]:  # type: ignore
        return BoundEqualTo


class NotEqualTo(LiteralPredicate):
    type: TypingLiteral["not-eq"] = Field(default="not-eq", alias="type")

    def __invert__(self) -> EqualTo:
        """Transform the Expression into its negated version."""
        return EqualTo(self.term, self.literal)

    @property
    def as_bound(self) -> type[BoundNotEqualTo]:  # type: ignore
        return BoundNotEqualTo


class LessThan(LiteralPredicate):
    type: TypingLiteral["lt"] = Field(default="lt", alias="type")

    def __invert__(self) -> GreaterThanOrEqual:
        """Transform the Expression into its negated version."""
        return GreaterThanOrEqual(self.term, self.literal)

    @property
    def as_bound(self) -> type[BoundLessThan]:  # type: ignore
        return BoundLessThan


class GreaterThanOrEqual(LiteralPredicate):
    type: TypingLiteral["gt-eq"] = Field(default="gt-eq", alias="type")

    def __invert__(self) -> LessThan:
        """Transform the Expression into its negated version."""
        return LessThan(self.term, self.literal)

    @property
    def as_bound(self) -> type[BoundGreaterThanOrEqual]:  # type: ignore
        return BoundGreaterThanOrEqual


class GreaterThan(LiteralPredicate):
    type: TypingLiteral["gt"] = Field(default="gt", alias="type")

    def __invert__(self) -> LessThanOrEqual:
        """Transform the Expression into its negated version."""
        return LessThanOrEqual(self.term, self.literal)

    @property
    def as_bound(self) -> type[BoundGreaterThan]:  # type: ignore
        return BoundGreaterThan


class LessThanOrEqual(LiteralPredicate):
    type: TypingLiteral["lt-eq"] = Field(default="lt-eq", alias="type")

    def __invert__(self) -> GreaterThan:
        """Transform the Expression into its negated version."""
        return GreaterThan(self.term, self.literal)

    @property
    def as_bound(self) -> type[BoundLessThanOrEqual]:  # type: ignore
        return BoundLessThanOrEqual


class StartsWith(LiteralPredicate):
    type: TypingLiteral["starts-with"] = Field(default="starts-with", alias="type")

    def __invert__(self) -> NotStartsWith:
        """Transform the Expression into its negated version."""
        return NotStartsWith(self.term, self.literal)

    @property
    def as_bound(self) -> type[BoundStartsWith]:  # type: ignore
        return BoundStartsWith


class NotStartsWith(LiteralPredicate):
    type: TypingLiteral["not-starts-with"] = Field(default="not-starts-with", alias="type")

    def __invert__(self) -> StartsWith:
        """Transform the Expression into its negated version."""
        return StartsWith(self.term, self.literal)

    @property
    def as_bound(self) -> type[BoundNotStartsWith]:  # type: ignore
        return BoundNotStartsWith



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import time
import warnings
from collections import defaultdict
from collections.abc import Iterable, Mapping
from enum import Enum
from typing import TYPE_CHECKING, Any

from pydantic import Field, PrivateAttr, model_serializer

from pyiceberg.io import FileIO
from pyiceberg.manifest import DataFile, DataFileContent, ManifestFile, _manifests
from pyiceberg.partitioning import UNPARTITIONED_PARTITION_SPEC, PartitionSpec
from pyiceberg.schema import Schema

if TYPE_CHECKING:
    from pyiceberg.table.metadata import TableMetadata
from pyiceberg.typedef import IcebergBaseModel

ADDED_DATA_FILES = "added-data-files"
ADDED_DELETE_FILES = "added-delete-files"
ADDED_EQUALITY_DELETES = "added-equality-deletes"
ADDED_FILE_SIZE = "added-files-size"
ADDED_POSITION_DELETES = "added-position-deletes"
ADDED_POSITION_DELETE_FILES = "added-position-delete-files"
ADDED_RECORDS = "added-records"
DELETED_DATA_FILES = "deleted-data-files"
DELETED_RECORDS = "deleted-records"
ADDED_EQUALITY_DELETE_FILES = "added-equality-delete-files"
REMOVED_DELETE_FILES = "removed-delete-files"
REMOVED_EQUALITY_DELETES = "removed-equality-deletes"
REMOVED_EQUALITY_DELETE_FILES = "removed-equality-delete-files"
REMOVED_FILE_SIZE = "removed-files-size"
REMOVED_POSITION_DELETES = "removed-position-deletes"
REMOVED_POSITION_DELETE_FILES = "removed-position-delete-files"
TOTAL_EQUALITY_DELETES = "total-equality-deletes"
TOTAL_POSITION_DELETES = "total-position-deletes"
TOTAL_DATA_FILES = "total-data-files"
TOTAL_DELETE_FILES = "total-delete-files"
TOTAL_RECORDS = "total-records"
TOTAL_FILE_SIZE = "total-files-size"
CHANGED_PARTITION_COUNT_PROP = "changed-partition-count"
CHANGED_PARTITION_PREFIX = "partitions."
PARTITION_SUMMARY_PROP = "partition-summaries-included"
OPERATION = "operation"

INITIAL_SEQUENCE_NUMBER = 0


class Operation(Enum):
    """Describes the operation.

    Possible operation values are:
        - append: Only data files were added and no files were removed.
        - replace: Data and delete files were added and removed without changing table data;
            i.e., compaction, changing the data file format, or relocating data files.
        - overwrite: Data and delete files were added and removed in a logical overwrite operation.
        - delete: Data files were removed and their contents logically deleted and/or delete files
            were added to delete rows.
    """

    APPEND = "append"
    REPLACE = "replace"
    OVERWRITE = "overwrite"
    DELETE = "delete"

    def __repr__(self) -> str:
        """Return the string representation of the Operation class."""
        return f"Operation.{self.name}"


class UpdateMetrics:
    added_file_size: int
    removed_file_size: int
    added_data_files: int
    removed_data_files: int
    added_eq_delete_files: int
    removed_eq_delete_files: int
    added_pos_delete_files: int
    removed_pos_delete_files: int
    added_delete_files: int
    removed_delete_files: int
    added_records: int
    deleted_records: int
    added_pos_deletes: int
    removed_pos_deletes: int
    added_eq_deletes: int
    removed_eq_deletes: int

    def __init__(self) -> None:
        self.added_file_size = 0
        self.removed_file_size = 0
        self.added_data_files = 0
        self.removed_data_files = 0
        self.added_eq_delete_files = 0
        self.removed_eq_delete_files = 0
        self.added_pos_delete_files = 0
        self.removed_pos_delete_files = 0
        self.added_delete_files = 0
        self.removed_delete_files = 0
        self.added_records = 0
        self.deleted_records = 0
        self.added_pos_deletes = 0
        self.removed_pos_deletes = 0
        self.added_eq_deletes = 0
        self.removed_eq_deletes = 0

    def add_file(self, data_file: DataFile) -> None:
        self.added_file_size += data_file.file_size_in_bytes

        if data_file.content == DataFileContent.DATA:
            self.added_data_files += 1
            self.added_records += data_file.record_count
        elif data_file.content == DataFileContent.POSITION_DELETES:
            self.added_delete_files += 1
            self.added_pos_delete_files += 1
            self.added_pos_deletes += data_file.record_count
        elif data_file.content == DataFileContent.EQUALITY_DELETES:
            self.added_delete_files += 1
            self.added_eq_delete_files += 1
            self.added_eq_deletes += data_file.record_count
        else:
            raise ValueError(f"Unknown data file content: {data_file.content}")

    def remove_file(self, data_file: DataFile) -> None:
        self.removed_file_size += data_file.file_size_in_bytes

        if data_file.content == DataFileContent.DATA:
            self.removed_data_files += 1
            self.deleted_records += data_file.record_count
        elif data_file.content == DataFileContent.POSITION_DELETES:
            self.removed_delete_files += 1
            self.removed_pos_delete_files += 1
            self.removed_pos_deletes += data_file.record_count
        elif data_file.content == DataFileContent.EQUALITY_DELETES:
            self.removed_delete_files += 1
            self.removed_eq_delete_files += 1
            self.removed_eq_deletes += data_file.record_count
        else:
            raise ValueError(f"Unknown data file content: {data_file.content}")

    def to_dict(self) -> dict[str, str]:
        properties: dict[str, str] = {}
        set_when_positive(properties, self.added_file_size, ADDED_FILE_SIZE)
        set_when_positive(properties, self.removed_file_size, REMOVED_FILE_SIZE)
        set_when_positive(properties, self.added_data_files, ADDED_DATA_FILES)
        set_when_positive(properties, self.removed_data_files, DELETED_DATA_FILES)
        set_when_positive(properties, self.added_eq_delete_files, ADDED_EQUALITY_DELETE_FILES)
        set_when_positive(properties, self.removed_eq_delete_files, REMOVED_EQUALITY_DELETE_FILES)
        set_when_positive(properties, self.added_pos_delete_files, ADDED_POSITION_DELETE_FILES)
        set_when_positive(properties, self.removed_pos_delete_files, REMOVED_POSITION_DELETE_FILES)
        set_when_positive(properties, self.added_delete_files, ADDED_DELETE_FILES)
        set_when_positive(properties, self.removed_delete_files, REMOVED_DELETE_FILES)
        set_when_positive(properties, self.added_records, ADDED_RECORDS)
        set_when_positive(properties, self.deleted_records, DELETED_RECORDS)
        set_when_positive(properties, self.added_pos_deletes, ADDED_POSITION_DELETES)
        set_when_positive(properties, self.removed_pos_deletes, REMOVED_POSITION_DELETES)
        set_when_positive(properties, self.added_eq_deletes, ADDED_EQUALITY_DELETES)
        set_when_positive(properties, self.removed_eq_deletes, REMOVED_EQUALITY_DELETES)
        return properties


class Summary(IcebergBaseModel, Mapping[str, str]):
    """A class that stores the summary information for a Snapshot.

    The snapshot summary’s operation field is used by some operations,
    like snapshot expiration, to skip processing certain snapshots.
    """

    operation: Operation = Field()
    _additional_properties: dict[str, str] = PrivateAttr()

    def __init__(self, operation: Operation | None = None, **data: Any) -> None:
        if operation is None:
            warnings.warn("Encountered invalid snapshot summary: operation is missing, defaulting to overwrite", stacklevel=2)
            operation = Operation.OVERWRITE
        super().__init__(operation=operation, **data)
        self._additional_properties = data

    def __getitem__(self, __key: str) -> Any | None:  # type: ignore
        """Return a key as it is a map."""
        if __key.lower() == "operation":
            return self.operation
        else:
            return self._additional_properties.get(__key)

    def __setitem__(self, key: str, value: Any) -> None:
        """Set a key as it is a map."""
        if key.lower() == "operation":
            self.operation = value
        else:
            self._additional_properties[key] = value

    def __len__(self) -> int:
        """Return the number of keys in the summary."""
        # Operation is required
        return 1 + len(self._additional_properties)

    @model_serializer
    def ser_model(self) -> dict[str, str]:
        return {
            "operation": str(self.operation.value),
            **self._additional_properties,
        }

    @property
    def additional_properties(self) -> dict[str, str]:
        return self._additional_properties

    def __repr__(self) -> str:
        """Return the string representation of the Summary class."""
        repr_properties = f", **{repr(self._additional_properties)}" if self._additional_properties else ""
        return f"Summary({repr(self.operation)}{repr_properties})"

    def __eq__(self, other: Any) -> bool:
        """Compare if the summary is equal to another summary."""
        return (
            self.operation == other.operation and self.additional_properties == other.additional_properties
            if isinstance(other, Summary)
            else False
        )


class Snapshot(IcebergBaseModel):
    snapshot_id: int = Field(alias="snapshot-id")
    parent_snapshot_id: int | None = Field(alias="parent-snapshot-id", default=None)
    sequence_number: int | None = Field(alias="sequence-number", default=INITIAL_SEQUENCE_NUMBER)
    timestamp_ms: int = Field(alias="timestamp-ms", default_factory=lambda: int(time.time() * 1000))
    manifest_list: str = Field(alias="manifest-list", description="Location of the snapshot's manifest list file")
    summary: Summary | None = Field(default=None)
    schema_id: int | None = Field(alias="schema-id", default=None)
    first_row_id: int | None = Field(
        alias="first-row-id", default=None, description="assigned to the first row in the first data file in the first manifest"
    )
    added_rows: int | None = Field(
        alias="added-rows", default=None, description="The upper bound of the number of rows with assigned row IDs"
    )

    def __str__(self) -> str:
        """Return the string representation of the Snapshot class."""
        operation = f"{self.summary.operation}: " if self.summary else ""
        parent_id = f", parent_id={self.parent_snapshot_id}" if self.parent_snapshot_id else ""
        schema_id = f", schema_id={self.schema_id}" if self.schema_id is not None else ""
        result_str = f"{operation}id={self.snapshot_id}{parent_id}{schema_id}"
        return result_str

    def __repr__(self) -> str:
        """Return the string representation of the Snapshot class."""
        fields = [
            f"snapshot_id={self.snapshot_id}",
            f"parent_snapshot_id={self.parent_snapshot_id}",
            f"sequence_number={self.sequence_number}",
            f"timestamp_ms={self.timestamp_ms}",
            f"manifest_list='{self.manifest_list}'",
            f"summary={repr(self.summary)}" if self.summary else None,
            f"schema_id={self.schema_id}" if self.schema_id is not None else None,
            f"first_row_id={self.first_row_id}" if self.first_row_id is not None else None,
            f"added_rows={self.added_rows}" if self.added_rows is not None else None,
        ]
        filtered_fields = [field for field in fields if field is not None]
        return f"Snapshot({', '.join(filtered_fields)})"

    def manifests(self, io: FileIO) -> list[ManifestFile]:
        """Return the manifests for the given snapshot."""
        return list(_manifests(io, self.manifest_list))


class MetadataLogEntry(IcebergBaseModel):
    metadata_file: str = Field(alias="metadata-file")
    timestamp_ms: int = Field(alias="timestamp-ms")


class SnapshotLogEntry(IcebergBaseModel):
    snapshot_id: int = Field(alias="snapshot-id")
    timestamp_ms: int = Field(alias="timestamp-ms")


class SnapshotSummaryCollector:
    metrics: UpdateMetrics
    partition_metrics: defaultdict[str, UpdateMetrics]
    max_changed_partitions_for_summaries: int

    def __init__(self, partition_summary_limit: int = 0) -> None:
        self.metrics = UpdateMetrics()
        self.partition_metrics = defaultdict(UpdateMetrics)
        self.max_changed_partitions_for_summaries = partition_summary_limit

    def set_partition_summary_limit(self, limit: int) -> None:
        self.max_changed_partitions_for_summaries = limit

    def add_file(self, data_file: DataFile, schema: Schema, partition_spec: PartitionSpec = UNPARTITIONED_PARTITION_SPEC) -> None:
        self.metrics.add_file(data_file)
        if len(data_file.partition) > 0:
            self.update_partition_metrics(partition_spec=partition_spec, file=data_file, is_add_file=True, schema=schema)

    def remove_file(
        self, data_file: DataFile, schema: Schema, partition_spec: PartitionSpec = UNPARTITIONED_PARTITION_SPEC
    ) -> None:
        self.metrics.remove_file(data_file)
        if len(data_file.partition) > 0:
            self.update_partition_metrics(partition_spec=partition_spec, file=data_file, is_add_file=False, schema=schema)

    def update_partition_metrics(self, partition_spec: PartitionSpec, file: DataFile, is_add_file: bool, schema: Schema) -> None:
        partition_path = partition_spec.partition_to_path(file.partition, schema)
        partition_metrics: UpdateMetrics = self.partition_metrics[partition_path]

        if is_add_file:
            partition_metrics.add_file(file)
        else:
            partition_metrics.remove_file(file)

    def build(self) -> dict[str, str]:
        properties = self.metrics.to_dict()
        changed_partitions_size = len(self.partition_metrics)
        set_when_positive(properties, changed_partitions_size, CHANGED_PARTITION_COUNT_PROP)
        if changed_partitions_size <= self.max_changed_partitions_for_summaries:
            if changed_partitions_size > 0:
                properties[PARTITION_SUMMARY_PROP] = "true"
            for partition_path, update_metrics_partition in self.partition_metrics.items():
                if (summary := self._partition_summary(update_metrics_partition)) and len(summary) != 0:
                    properties[CHANGED_PARTITION_PREFIX + partition_path] = summary

        return properties

    def _partition_summary(self, update_metrics: UpdateMetrics) -> str:
        return ",".join([f"{prop}={val}" for prop, val in update_metrics.to_dict().items()])


def update_snapshot_summaries(summary: Summary, previous_summary: Mapping[str, str] | None = None) -> Summary:
    if summary.operation not in {Operation.APPEND, Operation.OVERWRITE, Operation.DELETE}:
        raise ValueError(f"Operation not implemented: {summary.operation}")

    if not previous_summary:
        previous_summary = {
            TOTAL_DATA_FILES: "0",
            TOTAL_DELETE_FILES: "0",
            TOTAL_RECORDS: "0",
            TOTAL_FILE_SIZE: "0",
            TOTAL_POSITION_DELETES: "0",
            TOTAL_EQUALITY_DELETES: "0",
        }

    def _update_totals(total_property: str, added_property: str, removed_property: str) -> None:
        if previous_total_str := previous_summary.get(total_property):
            try:
                new_total = int(previous_total_str)
                if new_total >= 0 and (added := summary.get(added_property)):
                    new_total += int(added)
                if new_total >= 0 and (removed := summary.get(removed_property)):
                    new_total -= int(removed)
            except ValueError as e:
                raise ValueError(f"Could not parse summary property {total_property} to an int: {previous_total_str}") from e

            if new_total >= 0:
                summary[total_property] = str(new_total)

    _update_totals(
        total_property=TOTAL_DATA_FILES,
        added_property=ADDED_DATA_FILES,
        removed_property=DELETED_DATA_FILES,
    )
    _update_totals(
        total_property=TOTAL_DELETE_FILES,
        added_property=ADDED_DELETE_FILES,
        removed_property=REMOVED_DELETE_FILES,
    )
    _update_totals(
        total_property=TOTAL_RECORDS,
        added_property=ADDED_RECORDS,
        removed_property=DELETED_RECORDS,
    )
    _update_totals(
        total_property=TOTAL_FILE_SIZE,
        added_property=ADDED_FILE_SIZE,
        removed_property=REMOVED_FILE_SIZE,
    )
    _update_totals(
        total_property=TOTAL_POSITION_DELETES,
        added_property=ADDED_POSITION_DELETES,
        removed_property=REMOVED_POSITION_DELETES,
    )
    _update_totals(
        total_property=TOTAL_EQUALITY_DELETES,
        added_property=ADDED_EQUALITY_DELETES,
        removed_property=REMOVED_EQUALITY_DELETES,
    )

    return summary


def set_when_positive(properties: dict[str, str], num: int, property_name: str) -> None:
    if num > 0:
        properties[property_name] = str(num)


def ancestors_of(current_snapshot: Snapshot | None, table_metadata: TableMetadata) -> Iterable[Snapshot]:
    """Get the ancestors of and including the given snapshot."""
    snapshot = current_snapshot
    while snapshot is not None:
        yield snapshot
        if snapshot.parent_snapshot_id is None:
            break
        snapshot = table_metadata.snapshot_by_id(snapshot.parent_snapshot_id)


def ancestors_between(from_snapshot: Snapshot | None, to_snapshot: Snapshot, table_metadata: TableMetadata) -> Iterable[Snapshot]:
    """Get the ancestors of and including the given snapshot between the to and from snapshots."""
    if from_snapshot is not None:
        for snapshot in ancestors_of(to_snapshot, table_metadata):
            yield snapshot
            if snapshot == from_snapshot:
                break
    else:
        yield from ancestors_of(to_snapshot, table_metadata)


def latest_ancestor_before_timestamp(table_metadata: TableMetadata, timestamp_ms: int) -> Snapshot | None:
    """Find the latest ancestor snapshot whose timestamp is before the provided timestamp.

    Args:
        table_metadata: The table metadata for a table
        timestamp_ms: lookup snapshots strictly before this timestamp

    Returns:
        The latest ancestor snapshot older than the timestamp, or None if not found.
    """
    result: Snapshot | None = None
    result_timestamp: int = 0

    for ancestor in ancestors_of(table_metadata.current_snapshot(), table_metadata):
        if timestamp_ms > ancestor.timestamp_ms > result_timestamp:
            result = ancestor
            result_timestamp = ancestor.timestamp_ms

    return result



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pyiceberg.expressions import (
    Reference,
)
from pyiceberg.partitioning import (
    INITIAL_PARTITION_SPEC_ID,
    PARTITION_FIELD_ID_START,
    PartitionField,
    PartitionSpec,
    _PartitionNameGenerator,
    _visit_partition_field,
)
from pyiceberg.schema import Schema
from pyiceberg.table.update import (
    AddPartitionSpecUpdate,
    AssertDefaultSpecId,
    AssertLastAssignedPartitionId,
    SetDefaultSpecUpdate,
    TableRequirement,
    TableUpdate,
    UpdatesAndRequirements,
    UpdateTableMetadata,
)
from pyiceberg.transforms import IdentityTransform, TimeTransform, Transform, VoidTransform, parse_transform

if TYPE_CHECKING:
    from pyiceberg.table import Transaction


class UpdateSpec(UpdateTableMetadata["UpdateSpec"]):
    _transaction: Transaction
    _name_to_field: dict[str, PartitionField] = {}
    _name_to_added_field: dict[str, PartitionField] = {}
    _transform_to_field: dict[tuple[int, str], PartitionField] = {}
    _transform_to_added_field: dict[tuple[int, str], PartitionField] = {}
    _renames: dict[str, str] = {}
    _added_time_fields: dict[int, PartitionField] = {}
    _case_sensitive: bool
    _adds: list[PartitionField]
    _deletes: set[int]
    _last_assigned_partition_id: int

    def __init__(self, transaction: Transaction, case_sensitive: bool = True) -> None:
        super().__init__(transaction)
        self._name_to_field = {field.name: field for field in transaction.table_metadata.spec().fields}
        self._name_to_added_field = {}
        self._transform_to_field = {
            (field.source_id, repr(field.transform)): field for field in transaction.table_metadata.spec().fields
        }
        self._transform_to_added_field = {}
        self._adds = []
        self._deletes = set()
        self._last_assigned_partition_id = transaction.table_metadata.last_partition_id or PARTITION_FIELD_ID_START - 1
        self._renames = {}
        self._transaction = transaction
        self._case_sensitive = case_sensitive
        self._added_time_fields = {}

    def add_field(
        self,
        source_column_name: str,
        transform: str | Transform[Any, Any],
        partition_field_name: str | None = None,
    ) -> UpdateSpec:
        ref = Reference(source_column_name)
        bound_ref = ref.bind(self._transaction.table_metadata.schema(), self._case_sensitive)
        if isinstance(transform, str):
            transform = parse_transform(transform)
        # verify transform can actually bind it
        output_type = bound_ref.field.field_type
        if not transform.can_transform(output_type):
            raise ValueError(f"{transform} cannot transform {output_type} values from {bound_ref.field.name}")

        transform_key = (bound_ref.field.field_id, repr(transform))
        existing_partition_field = self._transform_to_field.get(transform_key)
        if existing_partition_field and self._is_duplicate_partition(transform, existing_partition_field):
            raise ValueError(f"Duplicate partition field for ${ref.name}=${ref}, ${existing_partition_field} already exists")

        added = self._transform_to_added_field.get(transform_key)
        if added:
            raise ValueError(f"Already added partition: {added.name}")

        new_field = self._partition_field((bound_ref.field.field_id, transform), partition_field_name)
        if new_field.name in self._name_to_added_field:
            raise ValueError(f"Already added partition field with name: {new_field.name}")

        if isinstance(new_field.transform, TimeTransform):
            existing_time_field = self._added_time_fields.get(new_field.source_id)
            if existing_time_field:
                raise ValueError(f"Cannot add time partition field: {new_field.name} conflicts with {existing_time_field.name}")
            self._added_time_fields[new_field.source_id] = new_field
        self._transform_to_added_field[transform_key] = new_field

        existing_partition_field = self._name_to_field.get(new_field.name)
        if existing_partition_field and new_field.field_id not in self._deletes:
            if isinstance(existing_partition_field.transform, VoidTransform):
                self.rename_field(
                    existing_partition_field.name, existing_partition_field.name + "_" + str(existing_partition_field.field_id)
                )
            else:
                raise ValueError(f"Cannot add duplicate partition field name: {existing_partition_field.name}")

        self._name_to_added_field[new_field.name] = new_field
        self._adds.append(new_field)
        return self

    def add_identity(self, source_column_name: str) -> UpdateSpec:
        return self.add_field(source_column_name, IdentityTransform(), None)

    def remove_field(self, name: str) -> UpdateSpec:
        added = self._name_to_added_field.get(name)
        if added:
            raise ValueError(f"Cannot delete newly added field {name}")
        renamed = self._renames.get(name)
        if renamed:
            raise ValueError(f"Cannot rename and delete field {name}")
        field = self._name_to_field.get(name)
        if not field:
            raise ValueError(f"No such partition field: {name}")

        self._deletes.add(field.field_id)
        return self

    def rename_field(self, name: str, new_name: str) -> UpdateSpec:
        existing_field = self._name_to_field.get(new_name)
        if existing_field and isinstance(existing_field.transform, VoidTransform):
            return self.rename_field(name, name + "_" + str(existing_field.field_id))
        added = self._name_to_added_field.get(name)
        if added:
            raise ValueError("Cannot rename recently added partitions")
        field = self._name_to_field.get(name)
        if not field:
            raise ValueError(f"Cannot find partition field {name}")
        if field.field_id in self._deletes:
            raise ValueError(f"Cannot delete and rename partition field {name}")
        self._renames[name] = new_name
        return self

    def _commit(self) -> UpdatesAndRequirements:
        new_spec = self._apply()
        updates: tuple[TableUpdate, ...] = ()
        requirements: tuple[TableRequirement, ...] = ()

        if self._transaction.table_metadata.default_spec_id != new_spec.spec_id:
            if new_spec.spec_id not in self._transaction.table_metadata.specs():
                updates = (
                    AddPartitionSpecUpdate(spec=new_spec),
                    SetDefaultSpecUpdate(spec_id=-1),
                )
            else:
                updates = (SetDefaultSpecUpdate(spec_id=new_spec.spec_id),)

            required_last_assigned_partitioned_id = self._transaction.table_metadata.last_partition_id
            default_spec_id = self._transaction.table_metadata.default_spec_id
            requirements = (
                AssertLastAssignedPartitionId(last_assigned_partition_id=required_last_assigned_partitioned_id),
                AssertDefaultSpecId(default_spec_id=default_spec_id),
            )

        return updates, requirements

    def _apply(self) -> PartitionSpec:
        def _check_and_add_partition_name(
            schema: Schema, name: str, source_id: int, transform: Transform[Any, Any], partition_names: set[str]
        ) -> None:
            from pyiceberg.partitioning import validate_partition_name

            validate_partition_name(name, transform, source_id, schema, partition_names)
            partition_names.add(name)

        def _add_new_field(
            schema: Schema, source_id: int, field_id: int, name: str, transform: Transform[Any, Any], partition_names: set[str]
        ) -> PartitionField:
            _check_and_add_partition_name(schema, name, source_id, transform, partition_names)
            return PartitionField(source_id, field_id, transform, name)

        partition_fields = []
        partition_names: set[str] = set()
        for field in self._transaction.table_metadata.spec().fields:
            if field.field_id not in self._deletes:
                renamed = self._renames.get(field.name)
                if renamed:
                    new_field = _add_new_field(
                        self._transaction.table_metadata.schema(),
                        field.source_id,
                        field.field_id,
                        renamed,
                        field.transform,
                        partition_names,
                    )
                else:
                    new_field = _add_new_field(
                        self._transaction.table_metadata.schema(),
                        field.source_id,
                        field.field_id,
                        field.name,
                        field.transform,
                        partition_names,
                    )
                partition_fields.append(new_field)
            elif self._transaction.table_metadata.format_version == 1:
                renamed = self._renames.get(field.name)
                if renamed:
                    new_field = _add_new_field(
                        self._transaction.table_metadata.schema(),
                        field.source_id,
                        field.field_id,
                        renamed,
                        VoidTransform(),
                        partition_names,
                    )
                else:
                    new_field = _add_new_field(
                        self._transaction.table_metadata.schema(),
                        field.source_id,
                        field.field_id,
                        field.name,
                        VoidTransform(),
                        partition_names,
                    )

                partition_fields.append(new_field)

        for added_field in self._adds:
            _check_and_add_partition_name(
                self._transaction.table_metadata.schema(),
                added_field.name,
                added_field.source_id,
                added_field.transform,
                partition_names,
            )
            new_field = PartitionField(
                source_id=added_field.source_id,
                field_id=added_field.field_id,
                transform=added_field.transform,
                name=added_field.name,
            )
            partition_fields.append(new_field)

        # Reuse spec id or create a new one.
        new_spec = PartitionSpec(*partition_fields)
        new_spec_id = INITIAL_PARTITION_SPEC_ID
        for spec in self._transaction.table_metadata.specs().values():
            if new_spec.compatible_with(spec):
                new_spec_id = spec.spec_id
                break
            elif new_spec_id <= spec.spec_id:
                new_spec_id = spec.spec_id + 1
        return PartitionSpec(*partition_fields, spec_id=new_spec_id)

    def _partition_field(self, transform_key: tuple[int, Transform[Any, Any]], name: str | None) -> PartitionField:
        if self._transaction.table_metadata.format_version == 2:
            source_id, transform = transform_key
            historical_fields = []
            for spec in self._transaction.table_metadata.specs().values():
                for field in spec.fields:
                    historical_fields.append(field)

            for field in historical_fields:
                if field.source_id == source_id and repr(field.transform) == repr(transform):
                    if name is None or field.name == name:
                        return PartitionField(source_id, field.field_id, transform, field.name)

        new_field_id = self._new_field_id()
        if name is None:
            tmp_field = PartitionField(transform_key[0], new_field_id, transform_key[1], "unassigned_field_name")
            name = _visit_partition_field(self._transaction.table_metadata.schema(), tmp_field, _PartitionNameGenerator())
        return PartitionField(transform_key[0], new_field_id, transform_key[1], name)

    def _new_field_id(self) -> int:
        self._last_assigned_partition_id += 1
        return self._last_assigned_partition_id

    def _is_duplicate_partition(self, transform: Transform[Any, Any], partition_field: PartitionField) -> bool:
        return partition_field.field_id not in self._deletes and partition_field.transform == transform



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import pytest

from pyiceberg.utils.properties import (
    get_first_property_value,
    property_as_bool,
    property_as_float,
    property_as_int,
)


def test_property_as_int() -> None:
    properties = {
        "int": "42",
    }

    assert property_as_int(properties, "int") == 42
    assert property_as_int(properties, "missing", default=1) == 1
    assert property_as_int(properties, "missing") is None


def test_property_as_int_with_invalid_value() -> None:
    properties = {
        "some_int_prop": "invalid",
    }

    with pytest.raises(ValueError) as exc:
        property_as_int(properties, "some_int_prop")

        assert "Could not parse table property some_int_prop to an integer: invalid" in str(exc.value)


def test_property_as_float() -> None:
    properties = {
        "float": "42.0",
    }

    assert property_as_float(properties, "float", default=1.0) == 42.0
    assert property_as_float(properties, "missing", default=1.0) == 1.0
    assert property_as_float(properties, "missing") is None


def test_property_as_float_with_invalid_value() -> None:
    properties = {
        "some_float_prop": "invalid",
    }

    with pytest.raises(ValueError) as exc:
        property_as_float(properties, "some_float_prop")

        assert "Could not parse table property some_float_prop to a float: invalid" in str(exc.value)


def test_property_as_bool() -> None:
    properties = {
        "bool": "True",
    }

    assert property_as_bool(properties, "bool", default=False) is True
    assert property_as_bool(properties, "missing", default=False) is False
    assert property_as_float(properties, "missing") is None


def test_property_as_bool_with_invalid_value() -> None:
    properties = {
        "some_bool_prop": "invalid",
    }

    with pytest.raises(ValueError) as exc:
        property_as_bool(properties, "some_bool_prop", True)

        assert "Could not parse table property some_bool_prop to a boolean: invalid" in str(exc.value)


def test_get_first_property_value() -> None:
    properties = {
        "prop_1": "value_1",
        "prop_2": "value_2",
    }

    assert get_first_property_value(properties, "prop_2", "prop_1") == "value_2"
    assert get_first_property_value(properties, "missing", "prop_1") == "value_1"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
# pylint:disable=redefined-outer-name
from datetime import date, datetime, timedelta, timezone
from decimal import Decimal
from typing import Any

import pytest
from pyspark.sql import SparkSession
from pyspark.sql.utils import AnalysisException

from pyiceberg.catalog import Catalog
from pyiceberg.partitioning import PartitionField, PartitionFieldValue, PartitionKey, PartitionSpec
from pyiceberg.schema import Schema
from pyiceberg.transforms import (
    BucketTransform,
    DayTransform,
    HourTransform,
    IdentityTransform,
    MonthTransform,
    TruncateTransform,
    YearTransform,
)
from pyiceberg.typedef import Record
from pyiceberg.types import (
    BinaryType,
    BooleanType,
    DateType,
    DecimalType,
    DoubleType,
    FixedType,
    FloatType,
    IntegerType,
    LongType,
    NestedField,
    StringType,
    TimestampType,
    TimestamptzType,
    UUIDType,
)

TABLE_SCHEMA = Schema(
    NestedField(field_id=1, name="boolean_field", field_type=BooleanType(), required=False),
    NestedField(field_id=2, name="string_field", field_type=StringType(), required=False),
    NestedField(field_id=3, name="string_long_field", field_type=StringType(), required=False),
    NestedField(field_id=4, name="int_field", field_type=IntegerType(), required=False),
    NestedField(field_id=5, name="long_field", field_type=LongType(), required=False),
    NestedField(field_id=6, name="float_field", field_type=FloatType(), required=False),
    NestedField(field_id=7, name="double_field", field_type=DoubleType(), required=False),
    NestedField(field_id=8, name="timestamp_field", field_type=TimestampType(), required=False),
    NestedField(field_id=9, name="timestamptz_field", field_type=TimestamptzType(), required=False),
    NestedField(field_id=10, name="date_field", field_type=DateType(), required=False),
    # NestedField(field_id=11, name="time", field_type=TimeType(), required=False),
    NestedField(field_id=11, name="binary_field", field_type=BinaryType(), required=False),
    NestedField(field_id=12, name="fixed_field", field_type=FixedType(16), required=False),
    NestedField(field_id=13, name="decimal_field", field_type=DecimalType(5, 2), required=False),
    NestedField(field_id=14, name="uuid_field", field_type=UUIDType(), required=False),
    NestedField(field_id=15, name="special#string+field", field_type=StringType(), required=False),
)


identifier = "default.test_table"


@pytest.mark.parametrize(
    (
        "partition_fields, partition_values, expected_partition_record, expected_hive_partition_path_slice, "
        "spark_create_table_sql_for_justification, spark_data_insert_sql_for_justification"
    ),
    [
        # Identity Transform
        (
            [PartitionField(source_id=1, field_id=1001, transform=IdentityTransform(), name="boolean_field")],
            [False],
            Record(False),
            "boolean_field=false",
            f"""CREATE TABLE {identifier} (
                boolean_field boolean,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                identity(boolean_field)  -- Partitioning by 'boolean_field'
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (false, 'Boolean field set to false');
            """,
        ),
        (
            [PartitionField(source_id=2, field_id=1001, transform=IdentityTransform(), name="string_field")],
            ["sample_string"],
            Record("sample_string"),
            "string_field=sample_string",
            f"""CREATE TABLE {identifier} (
                string_field string,
                another_string_field string
            )
            USING iceberg
            PARTITIONED BY (
                identity(string_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            ('sample_string', 'Another string value')
            """,
        ),
        (
            [PartitionField(source_id=4, field_id=1001, transform=IdentityTransform(), name="int_field")],
            [42],
            Record(42),
            "int_field=42",
            f"""CREATE TABLE {identifier} (
                int_field int,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                identity(int_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (42, 'Associated string value for int 42')
            """,
        ),
        (
            [PartitionField(source_id=5, field_id=1001, transform=IdentityTransform(), name="long_field")],
            [1234567890123456789],
            Record(1234567890123456789),
            "long_field=1234567890123456789",
            f"""CREATE TABLE {identifier} (
                long_field bigint,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                identity(long_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (1234567890123456789, 'Associated string value for long 1234567890123456789')
            """,
        ),
        (
            [PartitionField(source_id=6, field_id=1001, transform=IdentityTransform(), name="float_field")],
            [3.14],
            Record(3.14),
            "float_field=3.14",
            # spark writes differently as pyiceberg, Record[float_field=3.140000104904175],
            # path:float_field=3.14 (Record has difference)
            # so justification (compare expected value with spark behavior) would fail.
            None,
            None,
            # f"""CREATE TABLE {identifier} (
            #     float_field float,
            #     string_field string
            # )
            # USING iceberg
            # PARTITIONED BY (
            #     identity(float_field)
            # )
            # """,
            # f"""INSERT INTO {identifier}
            # VALUES
            # (3.14, 'Associated string value for float 3.14')
            # """
        ),
        (
            [PartitionField(source_id=7, field_id=1001, transform=IdentityTransform(), name="double_field")],
            [6.282],
            Record(6.282),
            "double_field=6.282",
            # spark writes differently as pyiceberg, Record[double_field=6.2820000648498535]
            # path:double_field=6.282 (Record has difference)
            # so justification (compare expected value with spark behavior) would fail.
            None,
            None,
            # f"""CREATE TABLE {identifier} (
            #     double_field double,
            #     string_field string
            # )
            # USING iceberg
            # PARTITIONED BY (
            #     identity(double_field)
            # )
            # """,
            # f"""INSERT INTO {identifier}
            # VALUES
            # (6.282, 'Associated string value for double 6.282')
            # """
        ),
        (
            [PartitionField(source_id=8, field_id=1001, transform=IdentityTransform(), name="timestamp_field")],
            [datetime(2023, 1, 1, 12, 0, 1, 999)],
            Record(1672574401000999),
            "timestamp_field=2023-01-01T12%3A00%3A01.000999",
            f"""CREATE TABLE {identifier} (
                timestamp_field timestamp_ntz,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                identity(timestamp_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01 12:00:01.000999' AS TIMESTAMP_NTZ), 'Associated string value for timestamp 2023-01-01T12:00:00')
            """,
        ),
        (
            [PartitionField(source_id=8, field_id=1001, transform=IdentityTransform(), name="timestamp_field")],
            [datetime(2023, 1, 1, 12, 0, 1)],
            Record(1672574401000000),
            "timestamp_field=2023-01-01T12%3A00%3A01",
            f"""CREATE TABLE {identifier} (
                timestamp_field timestamp_ntz,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                identity(timestamp_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01 12:00:01' AS TIMESTAMP_NTZ), 'Associated string value for timestamp 2023-01-01T12:00:00')
            """,
        ),
        (
            [PartitionField(source_id=8, field_id=1001, transform=IdentityTransform(), name="timestamp_field")],
            [datetime(2023, 1, 1, 12, 0, 0)],
            Record(1672574400000000),
            "timestamp_field=2023-01-01T12%3A00%3A00",
            # Spark writes differently as pyiceberg, so justification (compare expected value with spark behavior) would fail
            # AssertionError: assert 'timestamp_field=2023-01-01T12%3A00%3A00' in
            #     's3://warehouse/default/test_table/data/timestamp_field=2023-01-01T12%3A00/...-00001.parquet'
            # TLDR: CAST('2023-01-01 12:00:00' AS TIMESTAMP_NTZ) becomes 2023-01-01T12:00 in the hive partition path
            # when spark writes it (without the seconds).
            None,
            None,
            # f"""CREATE TABLE {identifier} (
            #     timestamp_field timestamp_ntz,
            #     string_field string
            # )
            # USING iceberg
            # PARTITIONED BY (
            #     identity(timestamp_field)
            # )
            # """,
            # f"""INSERT INTO {identifier}
            # VALUES
            # (CAST('2023-01-01 12:00:00' AS TIMESTAMP_NTZ), 'Associated string value for timestamp 2023-01-01T12:00:00')
            # """
        ),
        (
            [PartitionField(source_id=9, field_id=1001, transform=IdentityTransform(), name="timestamptz_field")],
            [datetime(2023, 1, 1, 12, 0, 1, 999, tzinfo=timezone(timedelta(hours=3)))],
            Record(1672563601000999),
            "timestamptz_field=2023-01-01T09%3A00%3A01.000999%2B00%3A00",
            # Spark writes differently as pyiceberg, so justification (compare expected value with spark behavior) would fail
            # AssertionError: assert 'timestamptz_field=2023-01-01T09%3A00%3A01.000999%2B00%3A00' in
            #     's3://warehouse/default/test_table/data/timestamptz_field=2023-01-01T09%3A00%3A01.000999Z/...-00001.parquet'
            # TLDR: CAST('2023-01-01 12:00:01.000999+03:00' AS TIMESTAMP) becomes 2023-01-01T09:00:01.000999Z in the hive
            # partition path when spark writes it (while iceberg: timestamptz_field=2023-01-01T09:00:01.000999+00:00).
            None,
            None,
            # f"""CREATE TABLE {identifier} (
            #     timestamptz_field timestamp,
            #     string_field string
            # )
            # USING iceberg
            # PARTITIONED BY (
            #     identity(timestamptz_field)
            # )
            # """,
            # f"""INSERT INTO {identifier}
            # VALUES
            # (CAST('2023-01-01 12:00:01.000999+03:00' AS TIMESTAMP),
            #     'Associated string value for timestamp 2023-01-01 12:00:01.000999+03:00')
            # """
        ),
        (
            [PartitionField(source_id=10, field_id=1001, transform=IdentityTransform(), name="date_field")],
            [date(2023, 1, 1)],
            Record(19358),
            "date_field=2023-01-01",
            f"""CREATE TABLE {identifier} (
                date_field date,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                identity(date_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01' AS DATE), 'Associated string value for date 2023-01-01')
            """,
        ),
        (
            [PartitionField(source_id=11, field_id=1001, transform=IdentityTransform(), name="binary_field")],
            [b"example"],
            Record(b"example"),
            "binary_field=ZXhhbXBsZQ%3D%3D",
            f"""CREATE TABLE {identifier} (
                binary_field binary,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                identity(binary_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('example' AS BINARY), 'Associated string value for binary `example`')
            """,
        ),
        (
            [PartitionField(source_id=13, field_id=1001, transform=IdentityTransform(), name="decimal_field")],
            [Decimal("123.45")],
            Record(Decimal("123.45")),
            "decimal_field=123.45",
            f"""CREATE TABLE {identifier} (
                decimal_field decimal(5,2),
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                identity(decimal_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (123.45, 'Associated string value for decimal 123.45')
            """,
        ),
        # # Year Month Day Hour Transform
        # Month Transform
        (
            [PartitionField(source_id=8, field_id=1001, transform=MonthTransform(), name="timestamp_field_month")],
            [datetime(2023, 1, 1, 11, 55, 59, 999999)],
            Record((2023 - 1970) * 12),
            "timestamp_field_month=2023-01",
            f"""CREATE TABLE {identifier} (
                timestamp_field timestamp_ntz,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                month(timestamp_field)  -- Partitioning by month from 'timestamp_field'
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01 11:55:59.999999' AS TIMESTAMP_NTZ), 'Event at 2023-01-01 11:55:59.999999');
            """,
        ),
        (
            [PartitionField(source_id=9, field_id=1001, transform=MonthTransform(), name="timestamptz_field_month")],
            [datetime(2023, 1, 1, 12, 0, 1, 999, tzinfo=timezone(timedelta(hours=3)))],
            Record((2023 - 1970) * 12 + 1 - 1),
            "timestamptz_field_month=2023-01",
            f"""CREATE TABLE {identifier} (
                timestamptz_field timestamp,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                month(timestamptz_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01 12:00:01.000999+03:00' AS TIMESTAMP), 'Event at 2023-01-01 12:00:01.000999+03:00');
            """,
        ),
        (
            [PartitionField(source_id=10, field_id=1001, transform=MonthTransform(), name="date_field_month")],
            [date(2023, 1, 1)],
            Record((2023 - 1970) * 12),
            "date_field_month=2023-01",
            f"""CREATE TABLE {identifier} (
                date_field date,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                month(date_field)  -- Partitioning by month from 'date_field'
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01' AS DATE), 'Event on 2023-01-01');
            """,
        ),
        # Year Transform
        (
            [PartitionField(source_id=8, field_id=1001, transform=YearTransform(), name="timestamp_field_year")],
            [datetime(2023, 1, 1, 11, 55, 59, 999999)],
            Record(2023 - 1970),
            "timestamp_field_year=2023",
            f"""CREATE TABLE {identifier}  (
                timestamp_field timestamp,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                year(timestamp_field)  -- Partitioning by year from 'timestamp_field'
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01 11:55:59.999999' AS TIMESTAMP), 'Event at 2023-01-01 11:55:59.999999');
            """,
        ),
        (
            [PartitionField(source_id=9, field_id=1001, transform=YearTransform(), name="timestamptz_field_year")],
            [datetime(2023, 1, 1, 12, 0, 1, 999, tzinfo=timezone(timedelta(hours=3)))],
            Record(53),
            "timestamptz_field_year=2023",
            f"""CREATE TABLE {identifier} (
                timestamptz_field timestamp,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                year(timestamptz_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01 12:00:01.000999+03:00' AS TIMESTAMP), 'Event at 2023-01-01 12:00:01.000999+03:00');
            """,
        ),
        (
            [PartitionField(source_id=10, field_id=1001, transform=YearTransform(), name="date_field_year")],
            [date(2023, 1, 1)],
            Record(2023 - 1970),
            "date_field_year=2023",
            f"""CREATE TABLE {identifier} (
                date_field date,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                year(date_field)  -- Partitioning by year from 'date_field'
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01' AS DATE), 'Event on 2023-01-01');
            """,
        ),
        # # Day Transform
        (
            [PartitionField(source_id=8, field_id=1001, transform=DayTransform(), name="timestamp_field_day")],
            [datetime(2023, 1, 1, 11, 55, 59, 999999)],
            Record(19358),
            "timestamp_field_day=2023-01-01",
            f"""CREATE TABLE {identifier} (
                timestamp_field timestamp,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                day(timestamp_field)  -- Partitioning by day from 'timestamp_field'
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01' AS DATE), 'Event on 2023-01-01');
            """,
        ),
        (
            [PartitionField(source_id=9, field_id=1001, transform=DayTransform(), name="timestamptz_field_day")],
            [datetime(2023, 1, 1, 12, 0, 1, 999, tzinfo=timezone(timedelta(hours=3)))],
            Record(19358),
            "timestamptz_field_day=2023-01-01",
            f"""CREATE TABLE {identifier} (
                timestamptz_field timestamp,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                day(timestamptz_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01 12:00:01.000999+03:00' AS TIMESTAMP), 'Event at 2023-01-01 12:00:01.000999+03:00');
            """,
        ),
        (
            [PartitionField(source_id=10, field_id=1001, transform=DayTransform(), name="date_field_day")],
            [date(2023, 1, 1)],
            Record(19358),
            "date_field_day=2023-01-01",
            f"""CREATE TABLE {identifier} (
                date_field date,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                day(date_field)  -- Partitioning by day from 'date_field'
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01' AS DATE), 'Event on 2023-01-01');
            """,
        ),
        # Hour Transform
        (
            [PartitionField(source_id=8, field_id=1001, transform=HourTransform(), name="timestamp_field_hour")],
            [datetime(2023, 1, 1, 11, 55, 59, 999999)],
            Record(464603),
            "timestamp_field_hour=2023-01-01-11",
            f"""CREATE TABLE {identifier} (
                timestamp_field timestamp,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                hour(timestamp_field)  -- Partitioning by hour from 'timestamp_field'
            )
            """,
            f"""INSERT INTO {identifier}
                VALUES
                (CAST('2023-01-01 11:55:59.999999' AS TIMESTAMP), 'Event within the 11th hour of 2023-01-01');
                """,
        ),
        (
            [PartitionField(source_id=9, field_id=1001, transform=HourTransform(), name="timestamptz_field_hour")],
            [datetime(2023, 1, 1, 12, 0, 1, 999, tzinfo=timezone(timedelta(hours=3)))],
            Record(464601),
            "timestamptz_field_hour=2023-01-01-09",
            f"""CREATE TABLE {identifier} (
                timestamptz_field timestamp,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                hour(timestamptz_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01 12:00:01.000999+03:00' AS TIMESTAMP), 'Event at 2023-01-01 12:00:01.000999+03:00');
            """,
        ),
        # Truncate Transform
        (
            [PartitionField(source_id=4, field_id=1001, transform=TruncateTransform(10), name="int_field_trunc")],
            [12345],
            Record(12340),
            "int_field_trunc=12340",
            f"""CREATE TABLE {identifier} (
                int_field int,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                truncate(int_field, 10)  -- Truncating 'int_field' integer column to a width of 10
            )
            """,
            f"""INSERT INTO {identifier}
                VALUES
                (12345, 'Sample data for int');
            """,
        ),
        (
            [PartitionField(source_id=5, field_id=1001, transform=TruncateTransform(2), name="bigint_field_trunc")],
            [2**32 + 1],
            Record(2**32),  # 4294967296
            "bigint_field_trunc=4294967296",
            f"""CREATE TABLE {identifier} (
                bigint_field bigint,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                truncate(bigint_field, 2)  -- Truncating 'bigint_field' long column to a width of 2
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (4294967297, 'Sample data for long');
            """,
        ),
        (
            [PartitionField(source_id=2, field_id=1001, transform=TruncateTransform(3), name="string_field_trunc")],
            ["abcdefg"],
            Record("abc"),
            "string_field_trunc=abc",
            f"""CREATE TABLE {identifier} (
                string_field string,
                another_string_field string
            )
            USING iceberg
            PARTITIONED BY (
                truncate(string_field, 3)  -- Truncating 'string_field' string column to a length of 3 characters
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            ('abcdefg', 'Another sample for string');
            """,
        ),
        (
            [PartitionField(source_id=13, field_id=1001, transform=TruncateTransform(width=5), name="decimal_field_trunc")],
            [Decimal("678.93")],
            Record(Decimal("678.90")),
            "decimal_field_trunc=678.90",  # Assuming truncation width of 1 leads to truncating to 670
            f"""CREATE TABLE {identifier} (
                decimal_field decimal(5,2),
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                truncate(decimal_field, 2)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (678.90, 'Associated string value for decimal 678.90')
            """,
        ),
        (
            [PartitionField(source_id=11, field_id=1001, transform=TruncateTransform(10), name="binary_field_trunc")],
            [b"HELLOICEBERG"],
            Record(b"HELLOICEBE"),
            "binary_field_trunc=SEVMTE9JQ0VCRQ%3D%3D",
            f"""CREATE TABLE {identifier} (
                binary_field binary,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                truncate(binary_field, 10)  -- Truncating 'binary_field' binary column to a length of 10 bytes
            )
            """,
            f"""INSERT INTO {identifier}
                VALUES
                (binary('HELLOICEBERG'), 'Sample data for binary');
            """,
        ),
        # Bucket Transform
        (
            [PartitionField(source_id=4, field_id=1001, transform=BucketTransform(2), name="int_field_bucket")],
            [10],
            Record(0),
            "int_field_bucket=0",
            f"""CREATE TABLE {identifier} (
                int_field int,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                bucket(2, int_field)  -- Distributing 'int_field' across 2 buckets
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (10, 'Integer with value 10');
            """,
        ),
        # Test multiple field combinations could generate the Partition record and hive partition path correctly
        (
            [
                PartitionField(source_id=8, field_id=1001, transform=YearTransform(), name="timestamp_field_year"),
                PartitionField(source_id=10, field_id=1002, transform=DayTransform(), name="date_field_day"),
            ],
            [
                datetime(2023, 1, 1, 11, 55, 59, 999999),
                date(2023, 1, 1),
            ],
            Record(53, 19358),
            "timestamp_field_year=2023/date_field_day=2023-01-01",
            f"""CREATE TABLE {identifier} (
                timestamp_field timestamp,
                date_field date,
                string_field string
            )
            USING iceberg
            PARTITIONED BY (
                year(timestamp_field),
                day(date_field)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            (CAST('2023-01-01 11:55:59.999999' AS TIMESTAMP), CAST('2023-01-01' AS DATE), 'some data');
            """,
        ),
        # Test that special characters are URL-encoded
        (
            [PartitionField(source_id=15, field_id=1001, transform=IdentityTransform(), name="special#string+field")],
            ["special string"],
            Record("special string"),
            "special%23string%2Bfield=special+string",
            f"""CREATE TABLE {identifier} (
                `special#string+field` string
            )
            USING iceberg
            PARTITIONED BY (
                identity(`special#string+field`)
            )
            """,
            f"""INSERT INTO {identifier}
            VALUES
            ('special string')
            """,
        ),
    ],
)
@pytest.mark.integration
def test_partition_key(
    session_catalog: Catalog,
    spark: SparkSession,
    partition_fields: list[PartitionField],
    partition_values: list[Any],
    expected_partition_record: Record,
    expected_hive_partition_path_slice: str,
    spark_create_table_sql_for_justification: str,
    spark_data_insert_sql_for_justification: str,
) -> None:
    field_values = [
        PartitionFieldValue(field, field.transform.transform(TABLE_SCHEMA.find_field(field.source_id).field_type)(value))
        for field, value in zip(partition_fields, partition_values, strict=True)
    ]
    spec = PartitionSpec(*partition_fields)

    key = PartitionKey(
        field_values=field_values,
        partition_spec=spec,
        schema=TABLE_SCHEMA,
    )

    # key.partition is used to write the metadata in DataFile, ManifestFile and all above layers
    assert key.partition == expected_partition_record
    # key.to_path() generates the hive partitioning part of the to-write parquet file path
    assert key.to_path() == expected_hive_partition_path_slice

    # Justify expected values are not made up but conforming to spark behaviors
    if spark_create_table_sql_for_justification is not None and spark_data_insert_sql_for_justification is not None:
        try:
            spark.sql(f"drop table {identifier}")
        except AnalysisException:
            pass

        spark.sql(spark_create_table_sql_for_justification)
        spark.sql(spark_data_insert_sql_for_justification)

        iceberg_table = session_catalog.load_table(identifier=identifier)
        snapshot = iceberg_table.current_snapshot()
        assert snapshot
        spark_partition_for_justification = (
            snapshot.manifests(iceberg_table.io)[0].fetch_manifest_entry(iceberg_table.io)[0].data_file.partition
        )
        spark_path_for_justification = (
            snapshot.manifests(iceberg_table.io)[0].fetch_manifest_entry(iceberg_table.io)[0].data_file.file_path
        )
        # Special characters in partition value are sanitized when written to the data file's partition field
        assert spark_partition_for_justification == expected_partition_record
        assert expected_hive_partition_path_slice in spark_path_for_justification



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import TYPE_CHECKING

import pytest

from pyiceberg.exceptions import NoSuchTableError
from pyiceberg.table.statistics import BlobMetadata, StatisticsFile

if TYPE_CHECKING:
    import pyarrow as pa

    from pyiceberg.catalog import Catalog
    from pyiceberg.schema import Schema
    from pyiceberg.table import Table


def _create_table_with_schema(catalog: "Catalog", schema: "Schema") -> "Table":
    tbl_name = "default.test_table_statistics_operations"

    try:
        catalog.drop_table(tbl_name)
    except NoSuchTableError:
        pass
    return catalog.create_table(identifier=tbl_name, schema=schema)


@pytest.mark.integration
@pytest.mark.parametrize("catalog", [pytest.lazy_fixture("session_catalog_hive"), pytest.lazy_fixture("session_catalog")])
def test_manage_statistics(catalog: "Catalog", arrow_table_with_null: "pa.Table") -> None:
    tbl = _create_table_with_schema(catalog, arrow_table_with_null.schema)

    tbl.append(arrow_table_with_null)
    tbl.append(arrow_table_with_null)

    add_snapshot_id_1 = tbl.history()[0].snapshot_id
    add_snapshot_id_2 = tbl.history()[1].snapshot_id

    def create_statistics_file(snapshot_id: int, type_name: str) -> StatisticsFile:
        blob_metadata = BlobMetadata(
            type=type_name,
            snapshot_id=snapshot_id,
            sequence_number=2,
            fields=[1],
            properties={"prop-key": "prop-value"},
        )

        statistics_file = StatisticsFile(
            snapshot_id=snapshot_id,
            statistics_path="s3://bucket/warehouse/stats.puffin",
            file_size_in_bytes=124,
            file_footer_size_in_bytes=27,
            blob_metadata=[blob_metadata],
        )

        return statistics_file

    statistics_file_snap_1 = create_statistics_file(add_snapshot_id_1, "apache-datasketches-theta-v1")
    statistics_file_snap_2 = create_statistics_file(add_snapshot_id_2, "deletion-vector-v1")

    with tbl.update_statistics() as update:
        update.set_statistics(statistics_file_snap_1)
        update.set_statistics(statistics_file_snap_2)

    assert len(tbl.metadata.statistics) == 2

    with tbl.update_statistics() as update:
        update.remove_statistics(add_snapshot_id_1)

    assert len(tbl.metadata.statistics) == 1

    with tbl.transaction() as txn:
        with txn.update_statistics() as update:
            update.set_statistics(statistics_file_snap_1)
            update.set_statistics(statistics_file_snap_2)

    assert len(tbl.metadata.statistics) == 2
